# SPDX-License-Identifier: Apache-2.0

import inspect
import pprint
import re
import sys
import traceback
import warnings
from logging import getLogger
import numpy as np
from scipy.sparse import coo_matrix
from onnx import SparseTensorProto, ValueInfoProto
from onnx.defs import onnx_opset_version, get_all_schemas_with_history
from onnx.helper import make_node, make_tensor, make_attribute, make_sparse_tensor
import onnx.onnx_cpp2py_export.defs as C
from onnxconverter_common.onnx_ops import __dict__ as dict_apply_operation
from ..proto import TensorProto
from .utils import get_domain


logger = getLogger("skl2onnx")


def _get_operation_list(use_shortlist=True):
    """
    Investigates this module to extract all ONNX functions
    which needs to be converted with these functions.
    """
    # Reduce the scope of method _check_operator,
    # it retrieves the stack trace and it takes a
    # significant amount of time.
    # This was mostly used to catch errors difficult to catch
    # otherwise.
    if use_shortlist:
        shortlist = {"Clip", "Normalizer", "Upsample"}
    else:
        shortlist = None
    regs = [
        re.compile("container.add_node[(]'([A-Z][a-zA-Z0-9]*)', " "\\[?input_name"),
        re.compile("container.add_node[(]'([A-Z][a-zA-Z0-9]*)', " "\\[\\]"),
        re.compile("container.add_node[(]'([A-Z][a-zA-Z0-9]*)', " "inputs"),
        re.compile("scope, '([A-Z][a-zA-Z0-9]*)', \\[?input_name"),
        re.compile("op_type = '([A-Z][a-zA-Z0-9]*)'"),
    ]
    res = {}
    for k, v in dict_apply_operation.items():
        if k.startswith("apply_") and callable(v):
            found = None
            source = inspect.getsource(v)
            for reg in regs:
                g = reg.search(source)
                if g:
                    found = g.groups()[0]
                    break
            if found is None:
                continue
            if shortlist and found not in shortlist:
                continue
            res[found] = v
    return res


def _build_options(model, defined_options, default_values, allowed_options, fail):
    opts = {} if default_values is None else default_values
    if defined_options is not None:
        opts.update(defined_options.get(type(model), {}))
        opts.update(defined_options.get(id(model), {}))
    if allowed_options not in (None, "passthrough"):
        for k, v in opts.items():
            if k not in allowed_options:
                if fail:
                    raise NameError(
                        "Option '{}' not in {} for class '{}'.".format(
                            k, list(sorted(allowed_options)), model.__class__.__name__
                        )
                    )
                return None
            allowed = allowed_options[k]
            if allowed is not None and v not in allowed and v is not None:
                raise ValueError(
                    "Unexpected value [{!r}] for option '{}'"
                    " (it must be in {}) for model '{}'.".format(
                        v, k, allowed, model.__class__.__name__
                    )
                )
    elif fail and len(opts) != 0 and allowed_options != "passthrough":
        raise RuntimeError(
            "Options {} are not registerd for model '{}'.".format(
                list(sorted(opts)), model.__class__.__name__
            )
        )
    return opts


_apply_operation_specific = _get_operation_list()


class _WhiteBlackContainer:
    def __init__(self, white_op=None, black_op=None, verbose=0):
        self._white_op = white_op
        self._black_op = black_op
        self.verbose = verbose

    def is_allowed(self, node_type):
        """
        Tells if a node is white listed or not black listed.
        """
        if isinstance(node_type, (list, tuple, set)):
            return all(map(self.is_allowed, node_type))
        try:
            self.check_white_black_list(node_type)
            return True
        except RuntimeError:
            return False

    def check_white_black_list(self, node_type):
        """
        Checks a node type is allowed according to white
        and black lists.
        """
        if self._white_op:
            if node_type not in self._white_op:
                raise RuntimeError(
                    "Operator '{}' is not white listed.".format(node_type)
                )
        if self._black_op:
            if node_type in self._black_op:
                raise RuntimeError("Operator '{}' is black listed.".format(node_type))

    def debug(self, *args, **kwargs):
        """
        Log debug information while converting a model.
        """
        logger.debug(*args, **kwargs)


class RawModelContainerNode(_WhiteBlackContainer):
    """
    This node is the carrier of the model we want to convert.
    It provides an abstract layer so that our parsing
    framework can work with models generated by different tools.
    """

    def __init__(self, raw_model, white_op=None, black_op=None, verbose=0):
        """
        :param raw_model: *scikit-learn* model to convert
        """
        _WhiteBlackContainer.__init__(
            self, white_op=white_op, black_op=black_op, verbose=verbose
        )
        self._raw_model = raw_model

    @property
    def raw_model(self):
        return self._raw_model

    @property
    def input_names(self):
        """
        This function should return a list of strings. Each string
        corresponds to an input variable name.
        :return: a list of string
        """
        raise NotImplementedError()

    @property
    def output_names(self):
        """
        This function should return a list of strings. Each string
        corresponds to an output variable name.
        :return: a list of string
        """
        raise NotImplementedError()


class SklearnModelContainerNode(RawModelContainerNode):
    """
    Main container for one *scikit-learn* model.
    Every converter adds nodes to an existing container
    which is converted into a *ONNX* graph by an instance of
    :class:`Topology <skl2onnx.common._topology.Topology>`.
    """

    def __init__(self, sklearn_model, white_op=None, black_op=None, verbose=0):
        super(SklearnModelContainerNode, self).__init__(
            sklearn_model, white_op=white_op, black_op=black_op, verbose=verbose
        )
        # Scikit-learn models have no input and output specified,
        # so we create them and store them in this container.
        self._inputs = []
        self._outputs = []

    @property
    def input_names(self):
        return [variable.onnx_name for variable in self._inputs]

    @property
    def output_names(self):
        return [variable.onnx_name for variable in self._outputs]

    def add_input(self, variable):
        # The order of adding variables matters. The final model's
        # input names are sequentially added as this list
        if variable not in self._inputs:
            self._inputs.append(variable)

    def add_output(self, variable):
        # The order of adding variables matters. The final model's
        # output names are sequentially added as this list
        if variable not in self._outputs:
            self._outputs.append(variable)


class ModelComponentContainer(_WhiteBlackContainer):
    """
    In the conversion phase, this class is used to collect all materials
    required to build an *ONNX* *GraphProto*, which is encapsulated in a
    *ONNX* *ModelProto*.
    """

    def __init__(
        self,
        target_opset,
        options=None,
        registered_models=None,
        white_op=None,
        black_op=None,
        verbose=0,
    ):
        """
        :param target_opset: number, for example, 7 for *ONNX 1.2*, and
                             8 for *ONNX 1.3*.
        :param options: see :ref:`l-conv-options`
        :param registered_models: registered models
        :param white_op: white list of ONNX nodes allowed
            while converting a pipeline, if empty, all are allowed
        :param black_op: black list of ONNX nodes allowed
            while converting a pipeline, if empty, none are blacklisted
        :param verbose: display information while converting
        """
        _WhiteBlackContainer.__init__(
            self, white_op=white_op, black_op=black_op, verbose=verbose
        )
        # Inputs of ONNX graph. They are ValueInfoProto in ONNX.
        self.inputs = []
        # Outputs of ONNX graph. They are ValueInfoProto in ONNX.
        self.outputs = []
        # ONNX tensors (type: TensorProto). They are initializers of
        # ONNX GraphProto.
        self.initializers = []
        self.initializers_strings = {}
        # Intermediate variables in ONNX computational graph. They are
        # ValueInfoProto in ONNX.
        self.value_info = []
        # ONNX nodes (type: NodeProto) used to define computation
        # structure
        self.nodes = []
        # ONNX operators' domain-version pair set. They will be added
        # into opset_import field in the final ONNX model.
        self.node_domain_version_pair_sets = set()
        # The targeted ONNX operator set (referred to as opset) that
        # matches the ONNX version.
        if isinstance(target_opset, dict):
            self.target_opset_all = target_opset
            self.target_opset = target_opset.get("", None)
        else:
            self.target_opset = target_opset
            self.target_opset_all = {"": target_opset}
        # Additional options given to converters.
        self.options = options
        # All registered models.
        self.registered_models = registered_models

    def swap_names(self, old_name, new_name):
        """
        Swaps variables names.

        :param old_name: old name
        :param new_name: new name
        :return: list of impacted objects
        """
        exc_list = {"Scan", "Loop", "If"}
        for node in self.nodes:
            if node.op_type not in exc_list:
                continue
            if (
                old_name in node.input
                or old_name in node.output
                or new_name in node.input
                or new_name in node.output
            ):
                raise NotImplementedError(
                    "Unable to handle subgraphs for node type %r."
                    "(%r, %r)" % (node.op_type, old_name, new_name)
                )
        res = []

        for inp in self.inputs:
            if inp.name == old_name:
                inp.name = new_name
                res.append(("Io", inp))
            elif inp.name == new_name:
                inp.name = old_name
                res.append(("In", inp))

        for inp in self.outputs:
            if inp.name == old_name:
                inp.name = new_name
                res.append(("Oo", inp))
            elif inp.name == new_name:
                inp.name = old_name
                res.append(("On", inp))

        for inp in self.initializers:
            if inp.name == old_name:
                inp.name = new_name
                res.append(("-o", inp))
            elif inp.name == new_name:
                inp.name = old_name
                res.append(("-n", inp))

        for node in self.nodes:
            modified = False
            new_input = []
            for name in node.input:
                if name == old_name:
                    name = new_name
                    modified = True
                elif name == new_name:
                    name = old_name
                    modified = True
                new_input.append(name)
            new_output = []
            for name in node.output:
                if name == old_name:
                    name = new_name
                    modified = True
                elif name == new_name:
                    name = old_name
                    modified = True
                new_output.append(name)
            if modified:
                if node.op_type in exc_list:
                    raise NotImplementedError(
                        "Unable to handle subgraphs for node type %r." "" % node.op_type
                    )
                node.input[:] = new_input[:]
                node.output[:] = new_output[:]
                res.append(("n-", node))
        return res

    def __str__(self):
        """
        Shows internal information.
        """
        rows = []
        if self.inputs:
            rows.append("INPUTS")
            for inp in self.inputs:
                rows.append("  " + str(inp).replace(" ", "").replace("\n", " "))
        if self.outputs:
            rows.append("OUTPUTS")
            for out in self.outputs:
                rows.append("  " + str(out).replace(" ", "").replace("\n", " "))
        if self.initializers:
            rows.append("INITIALIZERS")
            for ini in self.initializers:
                rows.append("  " + str(ini).replace(" ", "").replace("\n", " "))
        if self.value_info:
            rows.append("NODES")
            for val in self.value_info:
                rows.append("  " + str(val).replace(" ", "").replace("\n", " "))
        if self.nodes:
            rows.append("PROTO")
            for nod in self.nodes:
                rows.append("  " + str(nod).replace(" ", "").replace("\n", " "))
        return "\n".join(rows)

    def _make_value_info(self, variable):
        value_info = ValueInfoProto()
        value_info.name = variable.full_name
        value_info.type.CopyFrom(variable.type.to_onnx_type())
        if variable.type.doc_string:
            value_info.doc_string = variable.type.doc_string
        return value_info

    def add_input(self, variable):
        """
        Adds our *Variable* object defined _parser.py into the the input
        list of the final ONNX model.

        :param variable: The Variable object to be added
        """
        self.inputs.append(self._make_value_info(variable))

    def add_output(self, variable):
        """
        Adds our *Variable* object defined *_parser.py* into the the
        output list of the final ONNX model.

        :param variable: The Variable object to be added
        """
        self.outputs.append(self._make_value_info(variable))

    def add_options(self, model_id, options):
        """
        Adds an option, for example,
        ``add_options(id(clr), {'raw_scores': True})``
        tells the converter associated to ``clr`` to
        use raw score instead of probabilities.

        :param model_id: class or ``id(instance)``
        :param options: dictionary with the new values
        """
        if options is None:
            return
        if self.options is None:
            self.options = {}
        if model_id not in self.options:
            self.options[model_id] = None
        if self.options[model_id] is None:
            self.options[model_id] = {}
        self.options[model_id].update(options)

    def add_initializer(self, name, onnx_type, shape, content):
        """
        Adds a *TensorProto* into the initializer list of the final
        ONNX model.

        :param name: Variable name in the produced ONNX model.
        :param onnx_type: Element types allowed in ONNX tensor, e.g.,
                          TensorProto.FLOAT and TensorProto.STRING.
        :param shape: Tensor shape, a list of integers.
        :param content: Flattened tensor values (i.e., a float list
                        or a float array).
        :return: created tensor
        """
        logger.debug("[Init] %r, %r, %r", name, onnx_type, shape)
        sparse_tensor = None
        tensor = None

        cached_value = None
        if isinstance(content, TensorProto):
            tensor = TensorProto()
            tensor.data_type = content.data_type
            tensor.name = name
            tensor.raw_data = content.raw_data
            tensor.dims.extend(content.dims)
        elif shape is None and isinstance(
            content,
            (
                np.float32,
                np.float64,
                np.int32,
                np.int64,
                float,
                np.int8,
                np.uint8,
                np.bool_,
                np.str_,
                str,
            ),
        ):
            tensor = make_tensor(name, onnx_type, [], [content])
        elif SparseTensorProto is not None and isinstance(content, SparseTensorProto):
            raise NotImplementedError("Not implemented yet.")
        elif shape is None:
            tensor = make_attribute(name, content)
        elif isinstance(content, coo_matrix):
            if SparseTensorProto is None:
                raise RuntimeError(
                    "Sparse matrices require SparseTensorProto. Update onnx."
                )
            values_tensor = make_tensor(
                name + "_v",
                data_type=onnx_type,
                dims=(len(content.data),),
                vals=content.data,
            )
            indices = [
                i * content.shape[1] + j for i, j in zip(content.row, content.col)
            ]
            indices_tensor = make_tensor(
                name=name + "_i",
                data_type=TensorProto.INT64,
                dims=(len(indices),),
                vals=indices,
            )
            dense_shape = list(content.shape)
            sparse_tensor = make_sparse_tensor(
                values_tensor, indices_tensor, dense_shape
            )

            # cached value: same without names
            values_tensor = make_tensor(
                "_v", data_type=onnx_type, dims=(len(content.data),), vals=content.data
            )
            indices_tensor = make_tensor(
                name="_i",
                data_type=TensorProto.INT64,
                dims=(len(indices),),
                vals=indices,
            )
            cached_value = make_sparse_tensor(
                values_tensor, indices_tensor, dense_shape
            )

        else:
            if any(d is None for d in shape):
                raise ValueError("Shape of initializer cannot contain None.")
            if hasattr(content, "dtype") and content.dtype in (bool, np.bool_):
                content = content.astype(np.int32)
            try:
                tensor = make_tensor(name, onnx_type, shape, content)
            except TypeError as e:
                raise TypeError(
                    "Unable to make a tensor name=%r "
                    "onnx_type=%r shape=%r content-type=%r."
                    % (name, onnx_type, shape, type(content))
                ) from e

        if tensor is not None:
            if cached_value is None:
                name = tensor.name
                tensor.name = "tensor"
                content = tensor.SerializeToString()
                tensor.name = name
            else:
                content = cached_value.SerializeToString()
            cached_name = self.initializers_strings.get(content, None)
            if cached_name is None:
                self.initializers_strings[content] = name
                self.initializers.append(tensor)
                return tensor

            self.add_node(
                "Identity",
                cached_name,
                name,
                op_version=self.target_opset,
                name=name + "_op",
            )
            return name

        if sparse_tensor is not None:
            content = cached_value.SerializeToString()
            cached_name = self.initializers_strings.get(content, None)
            if cached_name is None:
                self.initializers_strings[content] = name
                self.add_node(
                    "Constant",
                    [],
                    [name],
                    sparse_value=sparse_tensor,
                    op_version=self.target_opset,
                    name=name + "_op",
                )
                return sparse_tensor

            self.add_node(
                "Identity",
                cached_name,
                name,
                op_version=self.target_opset,
                name=name + "_op",
            )
            return name

        raise RuntimeError("Either tensor or sparse_tensor should be defined.")

    def add_value_info(self, variable):
        self.value_info.append(self._make_value_info(variable))

    def _check_operator(self, op_type):
        """
        Checks that if *op_type* is one of the operators defined in
        :mod:`skl2onnx.common._apply_container`, then it was called
        from a function defined in this submodule by looking
        into the callstack. The test is enabled for *python >= 3.6*.
        """
        if op_type in _apply_operation_specific and sys.version_info[:2] >= (3, 6):
            tb = traceback.extract_stack()
            operation = []
            fct = _apply_operation_specific[op_type]
            skl2 = False
            for b in tb:
                if "_apply_operation" in b.filename and b.name == fct.__name__:
                    operation.append(b)
                    if not skl2 and "skl2onnx" in b.filename:
                        skl2 = True
            if skl2 and len(operation) == 0:
                raise RuntimeError(
                    "Operator '{0}' should be added with function "
                    "'{1}' in submodule _apply_operation.".format(op_type, fct.__name__)
                )
        self.check_white_black_list(op_type)

    def add_node(
        self,
        op_type,
        inputs,
        outputs,
        op_domain="",
        op_version=None,
        name=None,
        **attrs,
    ):
        """
        Adds a *NodeProto* into the node list of the final ONNX model.
        If the input operator's domain-version information cannot be
        found in our domain-version pool (a Python set), we may add it.

        :param op_type: A string (e.g., Pool and Conv) indicating the
                        type of the NodeProto
        :param inputs: A list of strings. They are the input variables'
                       names of the considered NodeProto
        :param outputs: A list of strings. They are the output
                        variables' names of the considered NodeProto
        :param op_domain: The domain name (e.g., ai.onnx.ml) of the
                          operator we are trying to add.
        :param op_version: The version number (e.g., 0 and 1) of the
                           operator we are trying to add.
        :param name: name of the node, this name cannot be empty
        :param attrs: A Python dictionary. Keys and values are
                      attributes' names and attributes' values,
                      respectively.
        """
        if "axes" in attrs and (
            attrs["axes"] is None or not isinstance(attrs["axes"], (list, np.ndarray))
        ):
            raise TypeError(
                f"axes must be a list or an array not {type(attrs['axes'])}."
            )
        if name is None or not isinstance(name, str) or name == "":
            name = f"N{len(self.nodes)}"
        existing_names = set(n.name for n in self.nodes)
        if name in existing_names:
            name += f"-N{len(self.nodes)}"

        if op_domain is None:
            op_domain = get_domain()
        self._check_operator(op_type)
        if op_version is None:
            op_version = self._get_op_version(op_domain, op_type)

        if isinstance(inputs, str):
            inputs = [inputs]
        if isinstance(outputs, str):
            outputs = [outputs]
        logger.debug(
            "[Node] %r - %r -> %r (name=%r)",
            op_type,
            ",".join(inputs),
            ",".join(outputs),
            name,
        )
        try:
            common = set(inputs) & set(outputs)
        except TypeError as e:
            raise TypeError(
                "inputs or outputs are wrong, inputs=%r, outputs=%r, node=%r."
                "" % (inputs, outputs, op_type)
            ) from e
        if common:
            raise RuntimeError(
                "inputs and outputs cannot have "
                "variables in common {} in node '{}' "
                "with name '{}'.".format(common, op_type, name)
            )
        if not isinstance(inputs, list) or not all(isinstance(s, str) for s in inputs):
            type_list = ",".join(list(str(type(s)) for s in inputs))
            raise ValueError("Inputs must be a list of string but get [%s]" % type_list)
        if not isinstance(outputs, list) or not all(
            isinstance(s, str) for s in outputs
        ):
            type_list = ",".join(list(str(type(s)) for s in outputs))
            raise ValueError(
                "Outputs must be a list of string but get [%s]" % type_list
            )
        upd = {}
        dtypes = set()
        for k, v in attrs.items():
            if v is None:
                raise ValueError(
                    "Failed to create ONNX node. Undefined "
                    "attribute pair (%s, %s) found for type %r and "
                    "version %r" % (k, v, op_type, op_version)
                )
            if isinstance(v, np.ndarray):
                upd[k] = v
                dtypes.add(v.dtype)

        if upd:
            attrs.update(upd)
        if "dtype" in attrs and op_type != "EyeLike":
            raise RuntimeError("dtype should not be a parameter.")

        new_attrs = {}
        for k, v in attrs.items():
            if hasattr(v, "shape") and len(v.shape) > 1:
                # handle type numpy.matrix
                new_attrs[k] = np.array(v).ravel()
                continue
            new_attrs[k] = v

        try:
            node = make_node(op_type, inputs, outputs, name=name, **new_attrs)
        except (ValueError, TypeError) as e:
            raise ValueError(
                f"Unable to create node {op_type!r} with name={name!r} and "
                f"attributes={pprint.pformat(new_attrs)}."
            ) from e
        node.domain = op_domain

        self.node_domain_version_pair_sets.add((op_domain, op_version))
        self.nodes.append(node)
        if (
            self.target_opset is not None
            and op_version is not None
            and op_version > self.target_opset_any_domain(op_domain)
        ):
            raise RuntimeError(
                "Opset number {} is higher than targeted opsets {} for "
                "node type '{}' name='{}' input={} "
                "output={} (domain='{}').".format(
                    op_version,
                    self.target_opset_all,
                    node.op_type,
                    node.name,
                    node.input,
                    node.output,
                    op_domain,
                )
            )

    def target_opset_any_domain(self, domain):
        target_opset = self.target_opset_all
        if isinstance(target_opset, dict):
            if domain in target_opset:
                to = target_opset[domain]
            else:
                to = None
            if to is None and domain == "":
                to = onnx_opset_version()
            if to is None:
                smap = C.schema_version_map()
                if domain in smap:
                    to = smap[domain][1]
            if to is not None:
                return to
            # The domain is not registered in onnx, it is probably
            # a custom domain. We assume the version is one.
            return 1
        return self.target_opset

    @property
    def target_opset_onnx(self):
        return self.target_opset_any_domain("")

    def _get_op_version(self, domain, op_type):
        """
        Determines the highest version of operator
        *op_type* below or equal to *target_opset*.
        """
        if not hasattr(self, "_op_versions"):
            self._build_op_version()
        key = domain, op_type
        vers = self._op_versions.get(key, None)
        if vers is None:
            if domain == "com.microsoft":
                # avoid a not necessarily necessary warning
                vers = 1
            else:
                warnings.warn(
                    "Unable to find operator '{}' in domain '{}' in ONNX, "
                    "op_version is forced to 1.".format(op_type, domain)
                )
            vers = [1]
        highest = self.target_opset_any_domain(domain)
        pos = len(vers) - 1
        while pos >= 0:
            if vers[pos] <= highest:
                return vers[pos]
            pos -= 1
        raise RuntimeError(
            "Unable to find a suitable version for operator '{}' "
            "in domain '{}'. Available versions: {}.".format(op_type, domain, vers)
        )

    def _build_op_version(self):
        res = {}
        for schema in get_all_schemas_with_history():
            dom = schema.domain
            name = schema.name
            vers = schema.since_version
            if (dom, name) not in res:
                res[dom, name] = set()
            res[dom, name].add(vers)
        self._op_versions = {}
        for k, v in res.items():
            self._op_versions[k] = list(sorted(v))

    def _get_allowed_options(self, model):
        if self.registered_models is not None:
            if inspect.isfunction(model):
                if model not in self.registered_models["aliases"]:
                    return None
                alias = self.registered_models["aliases"][model]
            elif hasattr(model, "alias"):
                alias = model.alias
            else:
                if type(model) not in self.registered_models["aliases"]:
                    return {}
                alias = self.registered_models["aliases"][type(model)]
            conv = self.registered_models["conv"][alias]
            allowed = conv.get_allowed_options()
            if allowed is None:
                return {}
            return allowed
        clname = str(model) if inspect.isfunction(model) else model.__class__.__name__
        raise NotImplementedError(
            "No registered models, no known allowed options "
            "for model '{}'.".format(clname)
        )

    def validate_options(self, operator):
        """
        Validates every operator allows the options
        given by the user at converter time
        for an operator.
        """
        skl_op = operator.raw_operator
        self.get_options(skl_op)

    def get_options(self, model, default_values=None, fail=True):
        """
        Returns additional options for a model.
        It first looks by class then by id (``id(model)``).
        :param model: model being converted
        :param default_values: default options (it is modified by
                               the function)
        :param fail: fails if options not found
        :return: dictionary
        """
        return _build_options(
            model,
            self.options,
            default_values,
            self._get_allowed_options(model),
            fail=fail,
        )

    def has_options(self, model, option_name):
        """
        Tells if a model allows one specific options.

        :param model: model being converted
        :return: boolean
        """
        opts = self._get_allowed_options(model)
        return option_name in opts

    def ensure_topological_order(self):
        """
        Ensures and modifies the order of nodes to have
        a topological order (every node in the list
        can only be an input for a node later in this list).
        The function raises an exception if a cycle is detected.
        """
        order = {}
        for inp in self.inputs:
            name = inp.name
            order[name] = 0
        for inp in self.initializers:
            name = inp.name
            order[name] = 0

        n_iter = 0
        missing_ops = []
        cont = True
        while cont and n_iter < len(self.nodes) * 2:
            n_iter += 1
            missing_names = set()
            missing_ops = []
            cont = False
            for node in self.nodes:
                maxi = 0
                for name in node.input:
                    if name in order:
                        maxi = max(maxi, order[name])
                    else:
                        maxi = None
                        missing_names.add(name)
                        break
                if maxi is None:
                    missing_ops.append(node)
                    continue
                key = id(node)
                if key in order:
                    continue
                cont = True
                maxi += 1
                order[key] = maxi
                maxi += 1
                for name in node.output:
                    if name in order:
                        raise RuntimeError(
                            "Unable to sort a node (cycle). An output was "
                            "already ordered with name %r (iteration=%r)."
                            "" % (name, n_iter)
                        )
                    order[name] = maxi
            if len(missing_names) == 0:
                continue

        if len(missing_ops) > 0:

            def nstr(name):
                if name in order:
                    return "%s#%d" % (name, order[name])
                return name

            rows = [
                "%s(%s) -> [%s]"
                % (
                    n.name or n.op_type,
                    ", ".join(map(nstr, n.input)),
                    ", ".join(n.output),
                )
                for n in missing_ops
            ]
            rows.insert(0, "")
            rows.append("--")
            rows.append("--all-nodes--")
            rows.append("--")
            rows.extend(
                "%s|%s(%s) -> [%s]"
                % (
                    n.op_type,
                    n.name or n.op_type,
                    ", ".join(map(nstr, n.input)),
                    ", ".join(n.output),
                )
                for n in self.nodes
            )
            raise RuntimeError(
                "After %d iterations for %d nodes, still unable "
                "to sort names %r. The graph may be disconnected. "
                "List of operators: %s"
                % (n_iter, len(self.nodes), missing_names, "\n".join(rows))
            )

        # Update order
        topo = sorted(
            [
                (order[id(node)], node.op_type, node.name, str(id(node)))
                for node in self.nodes
            ]
        )
        map_nodes = {str(id(node)): node for node in self.nodes}
        self.nodes = [map_nodes[_[-1]] for _ in topo]



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatType,
    Int64Type,
    StringType,
    TensorType,
    DoubleType,
    BooleanTensorType,
    FloatTensorType,
    Int64TensorType,
    StringTensorType,
    DoubleTensorType,
)
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_concat(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    for i in range(len(operator.inputs)):
        if len(operator.inputs[i].type.shape) != 2:
            operator.outputs[0].type.shape = [None, None]
            return

    N = operator.inputs[0].get_first_dimension()
    C = 0
    seen_types = []
    for i in operator.inputs:
        if C is not None:
            if isinstance(i.type, TensorType):
                if i.type.shape[1] is None:
                    C = None
                else:
                    C += i.type.shape[1]
            elif isinstance(i.type, (Int64Type, FloatType, StringType, DoubleType)):
                C += 1
            else:
                C = None
        if i.type not in seen_types:
            seen_types.append(i.type)

    def more_generic(t1, t2):
        if isinstance(t1, TensorType):
            if not isinstance(t2, TensorType):
                raise RuntimeError(
                    "Cannot merge columns with types {} and {}."
                    "Inputs:\n{}\nOutputs:\n{}".format(
                        t1, t2, operator.inputs, operator.outputs
                    )
                )
            for ts in [
                StringTensorType,
                DoubleTensorType,
                FloatTensorType,
                Int64TensorType,
                BooleanTensorType,
            ]:
                if isinstance(t1, ts) or isinstance(t2, ts):
                    return ts
            raise RuntimeError(
                "Cannot merge columns with types {} and {}."
                "Inputs:\n{}\nOutputs:\n{}".format(
                    t1, t2, operator.inputs, operator.outputs
                )
            )
        raise NotImplementedError(
            "Columns must be tensors."
            "Inputs:\n{}\nOutputs:\n{}".format(operator.inputs, operator.outputs)
        )

    # Let's determine the resulting type
    final_type = None
    for seen in seen_types:
        if final_type is None:
            final_type = seen
        elif seen.to_onnx_type() != final_type.to_onnx_type():
            merged_type = more_generic(final_type, seen)
            if isinstance(seen, merged_type):
                final_type = seen

    if final_type is None:
        raise NotImplementedError(
            "Columns must be tensors.\n"
            "- Inputs: {}\n- Outputs: {}\n- types: {}"
            "".format(operator.inputs, operator.outputs, seen_types)
        )
    if final_type != operator.outputs[0].type:
        operator.outputs[0].type = type(final_type)([N, C])
    else:
        operator.outputs[0].type.shape = [N, C]


register_shape_calculator("SklearnConcat", calculate_sklearn_concat)
register_shape_calculator("SklearnGenericUnivariateSelect", calculate_sklearn_concat)
register_shape_calculator("SklearnRFE", calculate_sklearn_concat)
register_shape_calculator("SklearnRFECV", calculate_sklearn_concat)
register_shape_calculator("SklearnSelectFdr", calculate_sklearn_concat)
register_shape_calculator("SklearnSelectFpr", calculate_sklearn_concat)
register_shape_calculator("SklearnSelectFromModel", calculate_sklearn_concat)
register_shape_calculator("SklearnSelectFwe", calculate_sklearn_concat)
register_shape_calculator("SklearnSelectKBest", calculate_sklearn_concat)
register_shape_calculator("SklearnSelectPercentile", calculate_sklearn_concat)
register_shape_calculator("SklearnVarianceThreshold", calculate_sklearn_concat)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_dict_vectorizer_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. Map ---> [1, C]

    C is the total number of allowed keys in the input dictionary.
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    C = len(operator.raw_operator.feature_names_)
    operator.outputs[0].type.shape = [None, C]


register_shape_calculator(
    "SklearnDictVectorizer", calculate_sklearn_dict_vectorizer_output_shapes
)



# SPDX-License-Identifier: Apache-2.0


import numbers
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_dict_vectorizer(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    When a *DictVectorizer* converts numbers into strings,
    scikit-learn adds a separator to disambiguate strings
    and still outputs floats. The method *predict*
    contains the following lines:

    ::

        if isinstance(v, str):
            f = "%s%s%s" % (f, self.separator, v)
            v = 1

    This cannot be implemented in ONNX. The converter
    raises an exception in that case.
    """
    op_type = "DictVectorizer"
    op = operator.raw_operator
    attrs = {"name": scope.get_unique_operator_name(op_type)}
    if all(isinstance(feature_name, str) for feature_name in op.feature_names_):
        # all strings, scikit-learn does the following:
        new_cats = []
        unique_cats = set()
        nbsep = 0
        for i in op.feature_names_:
            if op.separator in i:
                nbsep += 1
            if i in unique_cats:
                raise RuntimeError("Duplicated category '{}'.".format(i))
            unique_cats.add(i)
            new_cats.append(i)
        if nbsep >= len(new_cats):
            raise RuntimeError(
                "All categories contain a separator '{}'. "
                "This case is not supported by the converter. "
                "The mapping must map to numbers not string.".format(op.separator)
            )
        attrs["string_vocabulary"] = new_cats
    elif all(
        isinstance(feature_name, numbers.Integral) for feature_name in op.feature_names_
    ):
        attrs["int64_vocabulary"] = list(int(i) for i in op.feature_names_)
    else:
        raise ValueError("Keys must be all integers or all strings.")

    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("SklearnDictVectorizer", convert_sklearn_dict_vectorizer)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from scipy.linalg import solve_triangular
from sklearn.gaussian_process.kernels import ConstantKernel as C, RBF

try:
    from sklearn.gaussian_process._gpc import LAMBDAS, COEFS
except ImportError:
    LAMBDAS, COEFS = None, None
from ..proto import onnx_proto
from ..common.data_types import guess_numpy_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..algebra.onnx_ops import (
    OnnxAdd,
    OnnxSqrt,
    OnnxMatMul,
    OnnxSub,
    OnnxReduceSumApi11,
    OnnxMul,
    OnnxMax,
    OnnxReshapeApi13,
    OnnxDiv,
    OnnxNot,
    OnnxReciprocal,
    OnnxCast,
    OnnxLess,
    OnnxPow,
    OnnxNeg,
    OnnxConcat,
    OnnxArrayFeatureExtractor,
    OnnxTranspose,
)
from ..algebra.custom_ops import OnnxSolve

try:
    from ..algebra.onnx_ops import OnnxConstantOfShape
except ImportError:
    OnnxConstantOfShape = None
try:
    from ..algebra.onnx_ops import OnnxErf
except ImportError:
    OnnxErf = None
try:
    from ..algebra.onnx_ops import OnnxEinsum
except ImportError:
    OnnxEinsum = None
from ._gp_kernels import convert_kernel_diag, convert_kernel, _zero_vector_of_size


def convert_gaussian_process_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    The method *predict* from class *GaussianProcessRegressor*
    may cache some results if it is called with parameter
    ``return_std=True`` or ``return_cov=True``. This converter
    needs to be called with theses options to enable
    the second results.
    See example :ref:`l-gpr-example` to see how to
    use this converter which does not behave exactly
    as the others.
    """
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    opv = container.target_opset
    if opv is None:
        raise RuntimeError("container.target_opset must not be None")
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32

    options = container.get_options(
        op, dict(return_cov=False, return_std=False, optim=None)
    )
    if hasattr(op, "kernel_") and op.kernel_ is not None:
        kernel = op.kernel_
    elif op.kernel is None:
        kernel = C(1.0, constant_value_bounds="fixed") * RBF(
            1.0, length_scale_bounds="fixed"
        )
    else:
        kernel = op.kernel

    if not hasattr(op, "X_train_") or op.X_train_ is None:
        out0 = _zero_vector_of_size(
            X, keepdims=1, output_names=out[:1], dtype=dtype, op_version=opv
        )

        outputs = [out0]
        if options["return_cov"]:
            outputs.append(
                convert_kernel(
                    kernel, X, output_names=out[1:], dtype=dtype, op_version=opv
                )
            )
        if options["return_std"]:
            outputs.append(
                OnnxSqrt(
                    convert_kernel_diag(kernel, X, dtype=dtype, op_version=opv),
                    output_names=out[1:],
                    op_version=opv,
                )
            )
    else:
        # Code scikit-learn
        # K_trans = self.kernel_(X, self.X_train_)
        # y_mean = K_trans.dot(self.alpha_)  # Line 4 (y_mean = f_star)
        # y_mean = self._y_train_mean + y_mean * self._y_train_std

        k_trans = convert_kernel(
            kernel,
            X,
            x_train=op.X_train_.astype(dtype),
            dtype=dtype,
            optim=options.get("optim", None),
            op_version=opv,
        )
        k_trans.set_onnx_name_prefix("kgpd")
        y_mean_b = OnnxMatMul(k_trans, op.alpha_.astype(dtype), op_version=opv)

        mean_y = op._y_train_mean.astype(dtype)
        if len(mean_y.shape) == 1:
            mean_y = mean_y.reshape(mean_y.shape + (1,))

        if not hasattr(op, "_y_train_std") or op._y_train_std == 1:
            if isinstance(y_mean_b, (np.float32, np.float64)):
                y_mean_b = np.array([y_mean_b])
            if isinstance(mean_y, (np.float32, np.float64)):
                mean_y = np.array([mean_y])
            y_mean = OnnxAdd(y_mean_b, mean_y, op_version=opv)
        else:
            # A bug was fixed in 0.23 and it changed
            # the predictions when return_std is True.
            # See https://github.com/scikit-learn/scikit-learn/pull/15782.
            # y_mean = self._y_train_std * y_mean + self._y_train_mean
            var_y = op._y_train_std.astype(dtype)
            if len(var_y.shape) == 1:
                var_y = var_y.reshape(var_y.shape + (1,))
            if isinstance(var_y, (np.float32, np.float64)):
                var_y = np.array([var_y])
            if isinstance(mean_y, (np.float32, np.float64)):
                mean_y = np.array([mean_y])
            y_mean = OnnxAdd(
                OnnxMul(y_mean_b, var_y, op_version=opv), mean_y, op_version=opv
            )

        y_mean.set_onnx_name_prefix("gpr")
        y_mean_reshaped = OnnxReshapeApi13(
            y_mean,
            np.array([-1, 1], dtype=np.int64),
            op_version=opv,
            output_names=out[:1],
        )
        outputs = [y_mean_reshaped]

        if options["return_cov"]:
            raise NotImplementedError()
        if options["return_std"]:
            if hasattr(op, "_K_inv") and op._K_inv is not None:
                # scikit-learn < 0.24.2
                _K_inv = op._K_inv
            else:
                # scikit-learn >= 0.24.2
                L_inv = solve_triangular(op.L_.T, np.eye(op.L_.shape[0]))
                _K_inv = L_inv.dot(L_inv.T)

            # y_var = self.kernel_.diag(X)
            y_var = convert_kernel_diag(
                kernel, X, dtype=dtype, optim=options.get("optim", None), op_version=opv
            )

            # y_var -= np.einsum("ij,ij->i",
            #       np.dot(K_trans, self._K_inv), K_trans)
            k_dot = OnnxMatMul(k_trans, _K_inv.astype(dtype), op_version=opv)
            ys_var = OnnxSub(
                y_var,
                OnnxReduceSumApi11(
                    OnnxMul(k_dot, k_trans, op_version=opv),
                    axes=[1],
                    keepdims=0,
                    op_version=opv,
                ),
                op_version=opv,
            )

            # y_var_negative = y_var < 0
            # if np.any(y_var_negative):
            #     y_var[y_var_negative] = 0.0
            ys0_var = OnnxMax(ys_var, np.array([0], dtype=dtype), op_version=opv)

            if hasattr(op, "_y_train_std") and op._y_train_std != 1:
                # y_var = y_var * self._y_train_std**2
                ys0_var = OnnxMul(ys0_var, var_y**2, op_version=opv)

            # var = np.sqrt(ys0_var)
            var = OnnxSqrt(ys0_var, output_names=out[1:], op_version=opv)
            var.set_onnx_name_prefix("gprv")
            outputs.append(var)

    for o in outputs:
        o.add_to(scope, container)


def convert_gaussian_process_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    The method *predict* from class *GaussianProcessClassifier*
    may cache some results if it is called with parameter
    ``return_std=True`` or ``return_cov=True``. This converter
    needs to be called with theses options to enable
    the second results.
    See example :ref:`l-gpr-example` to see how to
    use this converter which does not behave exactly
    as the others.
    """
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    op_est = operator.raw_operator.base_estimator_
    opv = container.target_opset
    if opv is None:
        raise RuntimeError("container.target_opset must not be None")
    if OnnxEinsum is None or OnnxErf is None:
        raise RuntimeError(
            "target opset must be >= 12 for operator 'einsum' and 'erf'."
        )
    if LAMBDAS is None:
        raise RuntimeError("Only scikit-learn>=0.22 is supported.")
    outputs = []

    options = container.get_options(op, dict(optim=None))
    if hasattr(op, "kernel_") and op.kernel_ is not None:
        kernel = op.kernel_
    elif op.kernel is None:
        kernel = C(1.0, constant_value_bounds="fixed") * RBF(
            1.0, length_scale_bounds="fixed"
        )
    else:
        kernel = op.kernel

    if not hasattr(op_est, "X_train_"):
        raise NotImplementedError("Only binary classification is iplemented.")
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32
    K_starT = convert_kernel(
        kernel,
        X,
        x_train=op_est.X_train_.astype(dtype),
        dtype=dtype,
        optim=options.get("optim", None),
        op_version=opv,
    )
    K_star = OnnxTranspose(K_starT, op_version=opv)
    K_star.set_onnx_name_prefix("kstar")

    # common
    # f_star = K_star.T.dot(self.y_train_ - self.pi_)
    f_star_right = (op_est.y_train_ - op_est.pi_).astype(dtype).reshape((-1, 1))
    f_star = OnnxMatMul(K_starT, f_star_right, op_version=opv)
    f_star.set_onnx_name_prefix("f_star")

    best = OnnxCast(
        OnnxNot(
            OnnxLess(f_star, np.array([0], dtype=dtype), op_version=opv), op_version=opv
        ),
        to=onnx_proto.TensorProto.INT64,
        op_version=opv,
    )
    classes = OnnxArrayFeatureExtractor(op.classes_.astype(np.int64), best)
    labels = OnnxTranspose(classes, op_version=opv, output_names=out[:1])
    labels.set_onnx_name_prefix("labels")
    outputs.append(labels)

    # predict_proba
    # a x = b, x = a^-1 b
    # v = solve(self.L_, self.W_sr_[:, np.newaxis] * K_star)  # Line 5
    v = OnnxSolve(
        op_est.L_.astype(dtype),
        OnnxMul(op_est.W_sr_[:, np.newaxis].astype(dtype), K_star, op_version=opv),
        op_version=opv,
    )
    v.set_onnx_name_prefix("solve")

    # var_f_star = self.kernel_.diag(X) - np.einsum("ij,ij->j", v, v)
    var_f_star_kernel = convert_kernel_diag(
        kernel, X, dtype=dtype, optim=options.get("optim", None), op_version=opv
    )
    var_f_star_kernel.set_onnx_name_prefix("diag")
    var_f_star = OnnxSub(
        var_f_star_kernel,
        OnnxEinsum(v, v, equation="ij,ij->j", op_version=opv),
        op_version=opv,
    )
    var_f_star.set_onnx_name_prefix("var_f_star")

    # alpha = 1 / (2 * var_f_star)
    alpha = OnnxReciprocal(
        OnnxMul(var_f_star, np.array([2], dtype=dtype), op_version=opv), op_version=opv
    )
    alpha.set_onnx_name_prefix("alpha")

    # gamma = LAMBDAS * f_star
    gamma = OnnxMul(
        LAMBDAS.astype(dtype),
        OnnxReshapeApi13(f_star, np.array([1, -1], dtype=np.int64), op_version=opv),
        op_version=opv,
    )
    gamma.set_onnx_name_prefix("gamma")

    # integrals = np.sqrt(np.pi / alpha) *
    #               erf(gamma * np.sqrt(alpha / (alpha + LAMBDAS**2))) /
    #               (2 * np.sqrt(var_f_star * 2 * np.pi))
    integrals_1 = OnnxSqrt(
        OnnxDiv(np.array([np.pi], dtype=dtype), alpha, op_version=opv), op_version=opv
    )
    integrals_1.set_onnx_name_prefix("int1")

    integrals_2_1 = OnnxAdd(
        alpha,
        OnnxPow(LAMBDAS.astype(dtype), np.array([2], dtype=dtype), op_version=opv),
        op_version=opv,
    )
    integrals_2_1.set_onnx_name_prefix("int21")

    integrals_2_2 = OnnxSqrt(
        OnnxDiv(alpha, integrals_2_1, op_version=opv), op_version=opv
    )
    integrals_2_2.set_onnx_name_prefix("int22")

    integrals_div = OnnxMul(
        np.array([2], dtype=dtype),
        OnnxSqrt(
            OnnxMul(
                OnnxMul(var_f_star, np.array([2], dtype=dtype), op_version=opv),
                np.array([np.pi], dtype=dtype),
                op_version=opv,
            ),
            op_version=opv,
        ),
        op_version=opv,
    )
    integrals_div.set_onnx_name_prefix("intdiv")

    integrals = OnnxMul(
        integrals_1,
        OnnxDiv(
            OnnxErf(OnnxMul(gamma, integrals_2_2, op_version=opv), op_version=opv),
            integrals_div,
            op_version=opv,
        ),
        op_version=opv,
    )
    integrals.set_onnx_name_prefix("integrals")

    # pi_star = (COEFS * integrals).sum(axis=0) + .5 * COEFS.sum()
    coef_sum = (0.5 * COEFS.sum()).astype(dtype)
    if not isinstance(coef_sum, np.ndarray):
        coef_sum = np.array([coef_sum])
    pi_star = OnnxAdd(
        OnnxReduceSumApi11(
            OnnxMul(COEFS.astype(dtype), integrals, op_version=opv),
            op_version=opv,
            axes=[0],
        ),
        coef_sum,
        op_version=opv,
    )
    pi_star.set_onnx_name_prefix("pi_star")

    pi_star = OnnxReshapeApi13(
        pi_star, np.array([-1, 1], dtype=np.int64), op_version=opv
    )
    pi_star.set_onnx_name_prefix("pi_star2")
    final = OnnxConcat(
        OnnxAdd(
            OnnxNeg(pi_star, op_version=opv), np.array([1], dtype=dtype), op_version=opv
        ),
        pi_star,
        op_version=opv,
        axis=1,
        output_names=out[1:2],
    )
    outputs.append(final)

    for o in outputs:
        o.add_to(scope, container)


if OnnxConstantOfShape is not None:
    register_converter(
        "SklearnGaussianProcessRegressor",
        convert_gaussian_process_regressor,
        options={
            "return_cov": [False, True],
            "return_std": [False, True],
            "optim": [None, "cdist"],
        },
    )

if OnnxEinsum is not None and OnnxErf is not None:
    register_converter(
        "SklearnGaussianProcessClassifier",
        convert_gaussian_process_classifier,
        options={
            "optim": [None, "cdist"],
            "nocl": [False, True],
            "output_class_labels": [False, True],
            "zipmap": [False, True],
        },
    )



# SPDX-License-Identifier: Apache-2.0


import copy
import numpy as np
from ..common._registration import register_shape_calculator
from ..common.shape_calculator import calculate_linear_classifier_output_shapes
from ..common.data_types import FloatTensorType, Int64TensorType, DoubleTensorType
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_neighbours_transformer(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType, DoubleTensorType]
    )

    N = operator.inputs[0].get_first_dimension()
    n_samples_fit = operator.raw_operator.n_samples_fit_
    output_type = (
        DoubleTensorType
        if isinstance(operator.inputs[0].type, DoubleTensorType)
        else FloatTensorType
    )
    operator.outputs[0].type = output_type([N, n_samples_fit])


def calculate_sklearn_nearest_neighbours(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=[1, 2]
    )
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType, DoubleTensorType]
    )

    N = operator.inputs[0].get_first_dimension()
    neighbours = operator.raw_operator.n_neighbors
    operator.outputs[0].type = Int64TensorType([N, neighbours])
    operator.outputs[1].type.shape = [N, neighbours]


def calculate_sklearn_nearest_neighbours_regressor(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=[1, 2]
    )
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType, DoubleTensorType]
    )

    N = operator.inputs[0].get_first_dimension()
    if (
        hasattr(operator.raw_operator, "_y")
        and len(np.squeeze(operator.raw_operator._y).shape) == 1
    ):
        C = 1
    else:
        C = operator.raw_operator._y.shape[-1]
    operator.outputs[0].type.shape = [N, C]


def calculate_sklearn_nca(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType, DoubleTensorType]
    )

    N = operator.inputs[0].get_first_dimension()
    output_type = (
        DoubleTensorType
        if isinstance(operator.inputs[0].type, DoubleTensorType)
        else FloatTensorType
    )
    n_components = operator.raw_operator.components_.shape[0]
    operator.outputs[0].type = output_type([N, n_components])


def calculate_sklearn_knn_imputer(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape = operator.inputs[0].type.shape


register_shape_calculator(
    "SklearnKNeighborsRegressor", calculate_sklearn_nearest_neighbours_regressor
)
register_shape_calculator(
    "SklearnRadiusNeighborsRegressor", calculate_sklearn_nearest_neighbours_regressor
)
register_shape_calculator(
    "SklearnKNeighborsClassifier", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnRadiusNeighborsClassifier", calculate_linear_classifier_output_shapes
)
register_shape_calculator("SklearnKNNImputer", calculate_sklearn_knn_imputer)
register_shape_calculator(
    "SklearnKNeighborsTransformer", calculate_sklearn_neighbours_transformer
)
register_shape_calculator(
    "SklearnNearestNeighbors", calculate_sklearn_nearest_neighbours
)
register_shape_calculator(
    "SklearnNeighborhoodComponentsAnalysis", calculate_sklearn_nca
)



# SPDX-License-Identifier: Apache-2.0

import textwrap
import onnx
import onnx.defs  # noqa
from onnx.defs import OpSchema


def _get_doc_template():
    try:
        from jinja2 import Template
    except ImportError:

        class Template:
            def __init__(self, *args):
                pass

            def render(self, **context):
                schemas = context["schemas"]
                rows = []
                for sch in schemas:
                    doc = sch.doc or ""
                    name = sch.name
                    if name is None:
                        raise RuntimeError("An operator must have a name.")
                    rows.extend([name, "=" * len(name), "", doc, ""])
                return "\n".join(rows)

    return Template(
        textwrap.dedent(
            """
        {% for sch in schemas %}

        {{format_name_with_domain(sch)}}
        {{'=' * len(format_name_with_domain(sch))}}

        **Version**

        *Onnx name:* `{{sch.name}} <{{build_doc_url(sch)}}{{sch.name}}>`_

        {% if sch.support_level == OpSchema.SupportType.EXPERIMENTAL %}
        No versioning maintained for experimental ops.
        {% else %}
        This version of the operator has been {% if
        sch.deprecated %}deprecated{% else %}available{% endif %} since
        version {{sch.since_version}}{% if
        sch.domain %} of domain {{sch.domain}}{% endif %}.
        {% if len(sch.versions) > 1 %}
        Other versions of this operator:
        {% for v in sch.version[:-1] %} {{v}} {% endfor %}
        {% endif %}
        {% endif %}

        **Summary**

        {{process_documentation(sch.doc)}}

        {% if sch.attributes %}
        **Attributes**

        {% for _, attr in sorted(sch.attributes.items()) %}* *{{attr.name}}*{%
          if attr.required %} (required){% endif %}: {{attr.description}} {%
          if attr.default_value %}Default value is
          ``{{str(attr.default_value).replace('\\n', ' ').strip()}}``{%
          endif %}
        {% endfor %}
        {% endif %}

        {% if sch.inputs %}
        **Inputs**

        {% if sch.min_input != sch.max_input %}Between {{sch.min_input
        }} and {{sch.max_input}} inputs.
        {% endif %}
        {% for ii, inp in enumerate(sch.inputs) %}
        * *{{getname(inp, ii)}}*{{format_option(inp)}}{{get_type_str(inp)}}: {{
        inp.description}}{% endfor %}
        {% endif %}

        {% if sch.outputs %}
        **Outputs**

        {% if sch.min_output != sch.max_output %}Between {{sch.min_output
        }} and {{sch.max_output}} outputs.
        {% endif %}
        {% for ii, out in enumerate(sch.outputs) %}
        * *{{getname(out, ii)}}*{{format_option(out)}}{{get_type_str(out)}}: {{
        out.description}}{% endfor %}
        {% endif %}

        {% if sch.type_constraints %}
        **Type Constraints**

        {% for ii, type_constraint in enumerate(sch.type_constraints)
        %}* {{getconstraint(type_constraint, ii)}}: {{
        type_constraint.description}}
        {% endfor %}
        {% endif %}

        {% endfor %}
    """
        )
    )


_template_operator = _get_doc_template()


def get_domain_list():
    """
    Returns the list of available domains.
    """
    return list(
        sorted(set(map(lambda s: s.domain, onnx.defs.get_all_schemas_with_history())))
    )


def get_rst_doc(op_name=None):
    """
    Returns a documentation in RST format
    for all :class:`OnnxOperator`.

    :param op_name: operator name of None for all
    :return: string

    The function relies on module *jinja2* or replaces it
    with a simple rendering if not present.
    """
    if op_name is None:
        schemas = onnx.defs.get_all_schemas_with_history()
    elif isinstance(op_name, str):
        schemas = [
            schema
            for schema in onnx.defs.get_all_schemas_with_history()
            if schema.name == op_name
        ]
        if len(schemas) > 1:
            raise RuntimeError(
                "Multiple operators have the same name '{}'.".format(op_name)
            )
    elif not isinstance(op_name, list):
        schemas = [op_name]
    if len(schemas) == 0:
        raise ValueError("Unable to find any operator with name '{}'.".format(op_name))

    # from onnx.backend.sample.ops import collect_sample_implementations
    # from onnx.backend.test.case import collect_snippets
    # SNIPPETS = collect_snippets()
    # SAMPLE_IMPLEMENTATIONS = collect_sample_implementations()
    def format_name_with_domain(sch):
        if sch.domain:
            return "{} ({})".format(sch.name, sch.domain)
        return sch.name

    def get_is_homogeneous(obj):
        try:
            return obj.is_homogeneous
        except AttributeError:
            try:
                return obj.isHomogeneous
            except AttributeError:
                return False

    def format_option(obj):
        opts = []
        if OpSchema.FormalParameterOption.Optional == obj.option:
            opts.append("optional")
        elif OpSchema.FormalParameterOption.Variadic == obj.option:
            opts.append("variadic")
        if get_is_homogeneous(obj):
            opts.append("heterogeneous")
        if opts:
            return " (%s)" % ", ".join(opts)
        return ""

    def getconstraint(const, ii):
        if const.type_param_str:
            name = const.type_param_str
        else:
            name = str(ii)
        if const.allowed_type_strs:
            name += " " + ", ".join(const.allowed_type_strs)
        return name

    def getname(obj, i):
        name = obj.name
        if len(name) == 0:
            return str(i)
        return name

    def process_documentation(doc):
        if doc is None:
            doc = ""
        doc = textwrap.dedent(doc)
        main_docs_url = "https://github.com/onnx/onnx/blob/master/"
        rep = {
            "[the doc](IR.md)": "`ONNX <{0}docs/IR.md>`_",
            "[the doc](Broadcasting.md)": (
                "`Broadcasting in ONNX <{0}docs/Broadcasting.md>`_"
            ),
            "<dl>": "",
            "</dl>": "",
            "<dt>": "* ",
            "<dd>": "  ",
            "</dt>": "",
            "</dd>": "",
            "<tt>": "``",
            "</tt>": "``",
            "<br>": "\n",
        }
        for k, v in rep.items():
            doc = doc.replace(k, v.format(main_docs_url))
        move = 0
        lines = []
        for line in doc.split("\n"):
            if line.startswith("```"):
                if move > 0:
                    move -= 4
                    lines.append("\n")
                else:
                    lines.append("::\n")
                    move += 4
            elif move > 0:
                lines.append(" " * move + line)
            else:
                lines.append(line)
        return "\n".join(lines)

    def build_doc_url(sch):
        doc_url = "https://github.com/onnx/onnx/blob/master/docs/Operators"
        if "ml" in sch.domain:
            doc_url += "-ml"
        doc_url += ".md"
        doc_url += "#"
        if sch.domain not in (None, "", "ai.onnx"):
            doc_url += sch.domain + "."
        return doc_url

    def get_type_str(inou):
        try:
            return inou.type_str
        except AttributeError:
            return inou.typeStr

    fnwd = format_name_with_domain
    tmpl = _template_operator
    docs = tmpl.render(
        schemas=schemas,
        OpSchema=OpSchema,
        len=len,
        getattr=getattr,
        sorted=sorted,
        format_option=format_option,
        getconstraint=getconstraint,
        getname=getname,
        enumerate=enumerate,
        format_name_with_domain=fnwd,
        process_documentation=process_documentation,
        build_doc_url=build_doc_url,
        str=str,
        get_type_str=get_type_str,
    )
    return docs


def _get_doc_template_sklearn():
    try:
        from jinja2 import Template
    except ImportError:

        class Template:
            def __init__(self, *args):
                pass

            def render(self, **context):
                schemas = context["schemas"]
                rows = []
                for sch in schemas:
                    doc = sch.doc or ""
                    name = sch.name
                    if name is None:
                        raise RuntimeError("An operator must have a name.")
                    rows.extend([name, "=" * len(name), "", doc, ""])
                return "\n".join(rows)

    return Template(
        textwrap.dedent(
            """
        {% for cl in classes %}

        .. _l-sklops-{{cl.__name__}}:

        {{cl.__name__}}
        {{'=' * len(cl.__name__)}}

        Corresponding :class:`OnnxSubGraphOperatorMixin
        <skl2onnx.algebra.onnx_subgraph_operator_mixin.
        OnnxSubGraphOperatorMixin>` for model
        **{{cl.operator_name}}**.

        * Shape calculator: *{{cl._fct_shape_calc.__name__}}*
        * Converter: *{{cl._fct_converter.__name__}}*

        {{format_doc(cl)}}

        {% endfor %}
    """
        )
    )


_template_operator_sklearn = _get_doc_template_sklearn()


def get_rst_doc_sklearn():
    """
    Returns a documentation in RST format
    for all :class:`OnnxSubGraphOperatorMixin`.

    :param op_name: operator name of None for all
    :return: string

    The function relies on module *jinja2* or replaces it
    with a simple rendering if not present.
    """

    def format_doc(cl):
        return "\n".join(cl.__doc__.split("\n")[1:])

    from .sklearn_ops import dynamic_class_creation_sklearn

    classes = dynamic_class_creation_sklearn()
    tmpl = _template_operator_sklearn
    values = [(k, v) for k, v in sorted(classes.items())]
    values = [_[1] for _ in values]
    docs = tmpl.render(len=len, classes=values, format_doc=format_doc)
    return docs



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers, check_input_and_output_types
from ..common.shape_calculator import calculate_linear_regressor_output_shapes
from ..common.data_types import (
    BooleanTensorType,
    DoubleTensorType,
    FloatTensorType,
    Int64TensorType,
)


def calculate_bayesian_ridge_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, 1]

    This operator produces a scalar prediction for every example in a
    batch. If the input batch size is N, the output shape may be
    [N, 1].
    """
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=[1, 2]
    )
    check_input_and_output_types(
        operator,
        good_input_types=[
            BooleanTensorType,
            DoubleTensorType,
            FloatTensorType,
            Int64TensorType,
        ],
    )

    inp0 = operator.inputs[0].type
    if isinstance(inp0, (FloatTensorType, DoubleTensorType)):
        cls_type = inp0.__class__
    else:
        cls_type = FloatTensorType

    N = operator.inputs[0].get_first_dimension()
    if (
        hasattr(operator.raw_operator, "coef_")
        and len(operator.raw_operator.coef_.shape) > 1
    ):
        operator.outputs[0].type = cls_type([N, operator.raw_operator.coef_.shape[1]])
    else:
        operator.outputs[0].type = cls_type([N, 1])

    if len(operator.inputs) == 2:
        # option return_std is True
        operator.outputs[1].type = cls_type([N, 1])


register_shape_calculator(
    "SklearnAdaBoostRegressor", calculate_linear_regressor_output_shapes
)
register_shape_calculator(
    "SklearnBaggingRegressor", calculate_linear_regressor_output_shapes
)
register_shape_calculator(
    "SklearnBayesianRidge", calculate_bayesian_ridge_output_shapes
)
register_shape_calculator(
    "SklearnLinearRegressor", calculate_linear_regressor_output_shapes
)
register_shape_calculator("SklearnLinearSVR", calculate_linear_regressor_output_shapes)
register_shape_calculator(
    "SklearnMLPRegressor", calculate_linear_regressor_output_shapes
)
register_shape_calculator(
    "SklearnPoissonRegressor", calculate_linear_regressor_output_shapes
)
register_shape_calculator(
    "SklearnRANSACRegressor", calculate_linear_regressor_output_shapes
)
register_shape_calculator(
    "SklearnStackingRegressor", calculate_linear_regressor_output_shapes
)
register_shape_calculator(
    "SklearnTweedieRegressor", calculate_linear_regressor_output_shapes
)
register_shape_calculator(
    "SklearnGammaRegressor", calculate_linear_regressor_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import numbers
import numpy as np
from scipy.sparse import isspmatrix
from sklearn.svm import SVC, NuSVC, SVR, NuSVR, OneClassSVM
from ..common._apply_operation import apply_cast
from ..common.data_types import BooleanTensorType, Int64TensorType, guess_proto_type
from ..common._registration import register_converter
from ..proto import onnx_proto
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer

try:
    from ..common._apply_operation import apply_less
except ImportError:
    # onnxconverter-common is too old
    apply_less = None


def convert_sklearn_svm_regressor(
    scope: Scope,
    operator: Operator,
    container: ModelComponentContainer,
    op_type="SVMRegressor",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    """
    Converter for model
    `SVR <https://scikit-learn.org/stable/modules/
    generated/sklearn.svm.SVR.html>`_,
    `NuSVR <https://scikit-learn.org/stable/modules/
    generated/sklearn.svm.NuSVR.html>`_,
    `OneClassSVM <https://scikit-learn.org/stable/
    modules/generated/sklearn.svm.OneClassSVM.html>`_.
    The converted model in ONNX produces the same results as the
    original model except when probability=False:
    *onnxruntime* and *scikit-learn* do not return the same raw
    scores. *scikit-learn* returns aggregated scores
    as a *matrix[N, C]* coming from `_ovr_decision_function
    <https://github.com/scikit-learn/scikit-learn/blob/master/
    sklearn/utils/multiclass.py#L402>`_. *onnxruntime* returns
    the raw score from *svm* algorithm as a *matrix[N, (C(C-1)/2]*.
    """
    svm_attrs = {"name": scope.get_unique_operator_name("SVM")}
    op = operator.raw_operator
    if isinstance(op.dual_coef_, np.ndarray):
        coef = op.dual_coef_.ravel()
    else:
        coef = op.dual_coef_
    intercept = op.intercept_
    if isinstance(op.support_vectors_, np.ndarray):
        support_vectors = op.support_vectors_.ravel()
    else:
        support_vectors = op.support_vectors_

    svm_attrs["kernel_type"] = op.kernel.upper()
    svm_attrs["kernel_params"] = [
        np.float32(_) for _ in [op._gamma, op.coef0, op.degree]
    ]
    if isspmatrix(support_vectors):
        svm_attrs["support_vectors"] = support_vectors.toarray().ravel()
    else:
        svm_attrs["support_vectors"] = support_vectors
    if isspmatrix(coef):
        svm_attrs["coefficients"] = coef.toarray().ravel()
    else:
        svm_attrs["coefficients"] = coef
    svm_attrs["rho"] = intercept.astype(np.float32)
    svm_attrs["coefficients"] = svm_attrs["coefficients"].astype(np.float32)
    svm_attrs["support_vectors"] = svm_attrs["support_vectors"].astype(np.float32)

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    if operator.type in ["SklearnSVR", "SklearnNuSVR"] or isinstance(op, (SVR, NuSVR)):
        svm_attrs["post_transform"] = "NONE"
        svm_attrs["n_supports"] = len(op.support_)

        input_name = operator.input_full_names
        if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
            cast_input_name = scope.get_unique_variable_name("cast_input")
            apply_cast(
                scope,
                operator.input_full_names,
                cast_input_name,
                container,
                to=proto_dtype,
            )
            input_name = cast_input_name

        svm_out = scope.get_unique_variable_name("SVM03")
        container.add_node(
            op_type,
            input_name,
            svm_out,
            op_domain=op_domain,
            op_version=op_version,
            **svm_attrs
        )
        apply_cast(
            scope, svm_out, operator.output_full_names, container, to=proto_dtype
        )
    elif operator.type in ["SklearnOneClassSVM"] or isinstance(op, OneClassSVM):
        svm_attrs["post_transform"] = "NONE"
        svm_attrs["n_supports"] = len(op.support_)

        input_name = operator.input_full_names
        if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
            cast_input_name = scope.get_unique_variable_name("cast_input")
            apply_cast(
                scope,
                operator.input_full_names,
                cast_input_name,
                container,
                to=proto_dtype,
            )
            input_name = cast_input_name

        svm_out0 = scope.get_unique_variable_name("SVMO1")
        container.add_node(
            op_type,
            input_name,
            svm_out0,
            op_domain=op_domain,
            op_version=op_version,
            **svm_attrs
        )

        svm_out = operator.output_full_names[1]
        apply_cast(scope, svm_out0, svm_out, container, to=proto_dtype)

        pred = scope.get_unique_variable_name("float_prediction")
        container.add_node("Sign", svm_out, pred, op_version=9)
        apply_cast(
            scope,
            pred,
            operator.output_full_names[0],
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:
        raise ValueError(
            "Unknown support vector machine model type found "
            "'{0}'.".format(operator.type)
        )


def convert_sklearn_svm_classifier(
    scope: Scope,
    operator: Operator,
    container: ModelComponentContainer,
    op_type="SVMClassifier",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    """
    Converter for model
    `SVC <https://scikit-learn.org/stable/modules/
    generated/sklearn.svm.SVC.html>`_,
    `NuSVC <https://scikit-learn.org/stable/modules/
    generated/sklearn.svm.NuSVC.html>`_.
    The converted model in ONNX produces the same results as the
    original model except when probability=False:
    *onnxruntime* and *scikit-learn* do not return the same raw
    scores. *scikit-learn* returns aggregated scores
    as a *matrix[N, C]* coming from `_ovr_decision_function
    <https://github.com/scikit-learn/scikit-learn/blob/master/
    sklearn/utils/multiclass.py#L402>`_. *onnxruntime* returns
    the raw score from *svm* algorithm as a *matrix[N, (C(C-1)/2]*.
    """
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    svm_attrs = {"name": scope.get_unique_operator_name("SVMc")}
    op = operator.raw_operator
    if isinstance(op.dual_coef_, np.ndarray):
        coef = op.dual_coef_.ravel()
    else:
        coef = op.dual_coef_
    intercept = op.intercept_
    if isinstance(op.support_vectors_, np.ndarray):
        support_vectors = op.support_vectors_.ravel()
    elif isspmatrix(op.support_vectors_):
        support_vectors = op.support_vectors_.toarray().ravel()
    else:
        support_vectors = op.support_vectors_

    svm_attrs["kernel_type"] = op.kernel.upper()
    svm_attrs["kernel_params"] = [float(_) for _ in [op._gamma, op.coef0, op.degree]]
    svm_attrs["support_vectors"] = support_vectors

    if (
        operator.type in ["SklearnSVC", "SklearnNuSVC"] or isinstance(op, (SVC, NuSVC))
    ) and len(op.classes_) == 2:
        if isspmatrix(coef):
            coef_dense = coef.toarray().ravel()
            svm_attrs["coefficients"] = -coef_dense
        else:
            svm_attrs["coefficients"] = -coef
        svm_attrs["rho"] = -intercept
    else:
        if isspmatrix(coef):
            svm_attrs["coefficients"] = coef.todense()
        else:
            svm_attrs["coefficients"] = coef
        svm_attrs["rho"] = intercept

    handles_ovr = False
    svm_attrs["coefficients"] = svm_attrs["coefficients"].astype(np.float32)
    svm_attrs["support_vectors"] = svm_attrs["support_vectors"].astype(np.float32)
    svm_attrs["rho"] = svm_attrs["rho"].astype(np.float32)

    options = container.get_options(op, dict(raw_scores=False))
    use_raw_scores = options["raw_scores"]

    if operator.type in ["SklearnSVC", "SklearnNuSVC"] or isinstance(op, (SVC, NuSVC)):
        if len(op.probA_) > 0:
            svm_attrs["prob_a"] = op.probA_.astype(np.float32)
        else:
            handles_ovr = True
        if len(op.probB_) > 0:
            svm_attrs["prob_b"] = op.probB_.astype(np.float32)

        if (
            hasattr(op, "decision_function_shape")
            and op.decision_function_shape == "ovr"
            and handles_ovr
            and len(op.classes_) > 2
        ):
            output_name = scope.get_unique_variable_name("before_ovr")
        elif len(op.classes_) == 2 and use_raw_scores:
            output_name = scope.get_unique_variable_name("raw_scores")
        else:
            output_name = operator.outputs[1].full_name

        svm_attrs["post_transform"] = "NONE"
        svm_attrs["vectors_per_class"] = op.n_support_.tolist()

        label_name = operator.outputs[0].full_name
        probability_tensor_name = output_name

        if all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in op.classes_):
            labels = [int(i) for i in op.classes_]
            svm_attrs["classlabels_ints"] = labels
        elif all(isinstance(i, str) for i in op.classes_):
            labels = [str(i) for i in op.classes_]
            svm_attrs["classlabels_strings"] = labels
        else:
            raise RuntimeError("Invalid class label type '%s'." % op.classes_)

        svm_out = scope.get_unique_variable_name("SVM02")
        container.add_node(
            op_type,
            operator.inputs[0].full_name,
            [label_name, svm_out],
            op_domain=op_domain,
            op_version=op_version,
            **svm_attrs
        )
        apply_cast(scope, svm_out, probability_tensor_name, container, to=proto_dtype)
        if len(op.classes_) == 2 and use_raw_scores:
            minus_one = scope.get_unique_variable_name("minus_one")
            container.add_initializer(minus_one, proto_dtype, [], [-1])
            container.add_node(
                "Mul",
                [output_name, minus_one],
                operator.outputs[1].full_name,
                name=scope.get_unique_operator_name("MulRawScores"),
            )
    else:
        raise ValueError(
            "Unknown support vector machine model type found "
            "'{0}'.".format(operator.type)
        )

    if (
        hasattr(op, "decision_function_shape")
        and op.decision_function_shape == "ovr"
        and handles_ovr
        and len(op.classes_) > 2
    ):
        # Applies _ovr_decision_function.
        # See https://github.com/scikit-learn/scikit-learn/blob/
        # master/sklearn/utils/multiclass.py#L407:
        # ::
        #     _ovr_decision_function(dec < 0, -dec, len(self.classes_))

        if apply_less is None:
            raise RuntimeError(
                "Function apply_less is missing. " "onnxconverter-common is too old."
            )

        cst0 = scope.get_unique_variable_name("cst0")
        negative = scope.get_unique_variable_name("negative")
        container.add_initializer(cst0, proto_dtype, [], [0])
        apply_less(scope, [output_name, cst0], negative, container)
        inegative = scope.get_unique_variable_name("inegative")
        apply_cast(scope, negative, inegative, container, to=proto_dtype)

        score_name = scope.get_unique_variable_name("neg")
        container.add_node("Neg", [output_name], score_name)

        #
        #     ...
        #     def _ovr_decision_function(predictions, confidences, n_classes):
        #
        #     n_samples = predictions.shape[0]
        #     votes = np.zeros((n_samples, n_classes))
        #     sum_of_confidences = np.zeros((n_samples, n_classes))
        #     k = 0
        #     for i in range(n_classes):
        #         for j in range(i + 1, n_classes):
        #             sum_of_confidences[:, i] -= confidences[:, k]
        #             sum_of_confidences[:, j] += confidences[:, k]
        #             votes[predictions[:, k] == 0, i] += 1
        #             votes[predictions[:, k] == 1, j] += 1
        #             k += 1
        #     transformed_confidences = (
        #         sum_of_confidences / (3 * (np.abs(sum_of_confidences) + 1)))
        #     return votes + transformed_confidences

        this_operator = scope.declare_local_operator("SklearnOVRDecisionFunction", op)

        cl_type = operator.inputs[0].type.__class__
        prob_sign = scope.declare_local_variable("prob_sign", cl_type())
        container.add_node("Identity", [inegative], [prob_sign.onnx_name])
        prob_score = scope.declare_local_variable("prob_sign", cl_type())
        container.add_node("Identity", [score_name], [prob_score.onnx_name])

        this_operator.inputs.append(prob_sign)
        this_operator.inputs.append(prob_score)

        ovr_name = scope.declare_local_variable("ovr_output", cl_type())
        this_operator.outputs.append(ovr_name)

        output_name = operator.outputs[1].full_name
        container.add_node("Identity", [ovr_name.onnx_name], [output_name])


register_converter("SklearnOneClassSVM", convert_sklearn_svm_regressor)
register_converter(
    "SklearnSVC",
    convert_sklearn_svm_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "raw_scores": [True, False],
    },
)
register_converter("SklearnSVR", convert_sklearn_svm_regressor)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer
from ..common._apply_operation import apply_identity
from ..common.data_types import FloatTensorType, DoubleTensorType, guess_proto_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from .._supported_operators import sklearn_operator_name_map


def convert_sklearn_tfidf_vectoriser(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for scikit-learn's TfidfVectoriser.
    """
    tfidf_op = operator.raw_operator
    op_type = sklearn_operator_name_map[CountVectorizer]
    cv_operator = scope.declare_local_operator(op_type, tfidf_op)
    cv_operator.inputs = operator.inputs
    columns = max(operator.raw_operator.vocabulary_.values()) + 1
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    if proto_dtype == onnx_proto.TensorProto.FLOAT:
        clr = FloatTensorType
    elif proto_dtype == onnx_proto.TensorProto.DOUBLE:
        clr = DoubleTensorType
    else:
        raise RuntimeError(
            "Unexpected dtype '{}'. Float or double expected.".format(proto_dtype)
        )
    cv_output_name = scope.declare_local_variable(
        "count_vec_output", clr([None, columns])
    )
    cv_operator.outputs.append(cv_output_name)

    op_type = sklearn_operator_name_map[TfidfTransformer]
    tfidf_operator = scope.declare_local_operator(op_type, tfidf_op)
    tfidf_operator.inputs.append(cv_output_name)
    tfidf_output_name = scope.declare_local_variable("tfidf_output", clr())
    tfidf_operator.outputs.append(tfidf_output_name)

    apply_identity(
        scope, tfidf_output_name.full_name, operator.outputs[0].full_name, container
    )


register_converter(
    "SklearnTfidfVectorizer",
    convert_sklearn_tfidf_vectoriser,
    options={
        "tokenexp": None,
        "separators": None,
        "nan": [True, False],
        "keep_empty_string": [True, False],
        "locale": None,
    },
)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..common.data_types import (
    StringTensorType,
    Int64TensorType,
    FloatTensorType,
    DoubleTensorType,
)
from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_feature_hasher(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[StringTensorType, Int64TensorType]
    )

    N = operator.inputs[0].get_first_dimension()
    model = operator.raw_operator
    shape = [N, model.n_features]
    if model.dtype == np.float32:
        operator.outputs[0].type = FloatTensorType(shape=shape)
    elif model.dtype == np.float64:
        operator.outputs[0].type = DoubleTensorType(shape=shape)
    elif model.dtype in (np.int32, np.uint32, np.int64):
        operator.outputs[0].type = Int64TensorType(shape=shape)
    else:
        raise RuntimeError(
            f"Converter is not implemented for FeatureHasher.dtype={model.dtype}."
        )


register_shape_calculator("SklearnFeatureHasher", calculate_sklearn_feature_hasher)



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import apply_mul
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_proto_type


def convert_sklearn_multiply(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    for input, output in zip(operator.inputs, operator.outputs):
        operand_name = scope.get_unique_variable_name("operand")

        container.add_initializer(
            operand_name, guess_proto_type(input.type), [], [operator.operand]
        )

        apply_mul(
            scope,
            [input.full_name, operand_name],
            output.full_name,
            container,
        )


register_converter("SklearnMultiply", convert_sklearn_multiply)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import FloatTensorType, DoubleTensorType
from ..common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_sklearn_kernel_pca_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType, DoubleTensorType],
        good_output_types=[FloatTensorType, DoubleTensorType],
    )
    N = operator.inputs[0].get_first_dimension()
    op = operator.raw_operator
    lbd = op.eigenvalues_ if hasattr(op, "eigenvalues_") else op.lambdas_
    C = lbd.shape[0]
    operator.outputs[0].type.shape = [N, C]


def calculate_sklearn_kernel_centerer_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType, DoubleTensorType],
        good_output_types=[FloatTensorType, DoubleTensorType],
    )
    N = operator.inputs[0].get_first_dimension()
    C = operator.raw_operator.K_fit_rows_.shape[0]
    operator.outputs[0].type.shape = [N, C]


register_shape_calculator(
    "SklearnKernelCenterer", calculate_sklearn_kernel_centerer_output_shapes
)
register_shape_calculator(
    "SklearnKernelPCA", calculate_sklearn_kernel_pca_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers
from ..common.shape_calculator import _infer_linear_classifier_output_types


def voting_classifier_shape_calculator(operator):
    check_input_and_output_numbers(operator, output_count_range=2)

    _infer_linear_classifier_output_types(operator)


register_shape_calculator("SklearnVotingClassifier", voting_classifier_shape_calculator)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import FloatTensorType, Int64TensorType, DoubleTensorType
from ..common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_pls_regression_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType, DoubleTensorType]
    )

    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError("Input must be a [N, C]-tensor")

    op = operator.raw_operator
    cls_type = operator.inputs[0].type.__class__
    if cls_type != DoubleTensorType:
        cls_type = FloatTensorType
    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = cls_type([N, op.coef_.shape[1]])


register_shape_calculator(
    "SklearnPLSRegression", calculate_pls_regression_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import warnings
import numpy as np
from onnx import TensorProto
from ..common._registration import register_converter
from ..common.data_types import (
    BooleanTensorType,
    Int64TensorType,
    guess_numpy_type,
    guess_proto_type,
)
from ..algebra.onnx_ops import (
    OnnxCast,
    OnnxLess,
    OnnxMul,
    OnnxAdd,
    OnnxDiv,
    OnnxGather,
    OnnxReduceMeanApi18,
    OnnxMax,
    OnnxSqueezeApi11,
)
from .nearest_neighbours import onnx_nearest_neighbors_indices_k


def convert_sklearn_local_outlier_factor(
    scope,
    operator,
    container,
    op_type="TreeEnsembleRegressor",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    op = operator.raw_operator
    if not op.novelty:
        raise RuntimeError(
            "The converter only converts the model %r is novelty is True." "" % type(op)
        )
    outputs = operator.outputs
    opv = container.target_opset
    options = container.get_options(op, dict(score_samples=None, optim=None))

    X = operator.inputs[0]
    dtype = guess_numpy_type(operator.inputs[0].type)
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        X = OnnxCast(X, to=proto_dtype, op_version=opv)

    metric = op.effective_metric_ if hasattr(op, "effective_metric_") else op.metric
    neighb = op._fit_X.astype(dtype)
    k = op.n_neighbors_
    kwargs = {}
    if op.p != 2:
        if options["optim"] == "cdist":
            warnings.warn(
                "Option p=%r may not be compatible with the runtime. "
                "See https://github.com/microsoft/onnxruntime/blob/master/"
                "docs/ContribOperators.md#com.microsoft.CDist."
            )
        kwargs["p"] = op.p

    top_k, dist = onnx_nearest_neighbors_indices_k(
        X,
        neighb,
        k,
        metric,
        dtype=dtype,
        op_version=opv,
        keep_distances=True,
        optim=options.get("optim", None),
        **kwargs
    )

    # dist_k = self._distances_fit_X_[neighbors_indices, self.n_neighbors_ - 1]
    # reach_dist_array = np.maximum(distances_X, dist_k)
    dist_k_ = OnnxGather(op._distances_fit_X_.astype(dtype), top_k, op_version=opv)
    dist_k = OnnxSqueezeApi11(
        OnnxGather(
            dist_k_,
            np.array([op.n_neighbors_ - 1], dtype=np.int64),
            axis=2,
            op_version=opv,
        ),
        axes=[2],
        op_version=opv,
    )
    dist_k.set_onnx_name_prefix("dist_k")
    reach_dist_array = OnnxMax(
        OnnxMul(dist, np.array([-1], dtype=dtype), op_version=opv),
        dist_k,
        op_version=opv,
    )

    # X_lrd=  return 1.0 / (np.mean(reach_dist_array, axis=1) + 1e-10)
    X_lrd = OnnxDiv(
        np.array([1], dtype=dtype),
        OnnxAdd(
            OnnxReduceMeanApi18(reach_dist_array, axes=[1], op_version=opv, keepdims=1),
            np.array([1e-10], dtype=dtype),
            op_version=opv,
        ),
        op_version=opv,
    )
    X_lrd.set_onnx_name_prefix("X_lrd")

    # lrd_ratios_array = self._lrd[neighbors_indices_X] / X_lrd[:, np.newaxis]
    lrd_ratios_array = OnnxDiv(
        OnnxGather(op._lrd.astype(dtype), top_k, op_version=opv), X_lrd, op_version=opv
    )
    lrd_ratios_array.set_onnx_name_prefix("lrd_ratios_array")

    # -np.mean(lrd_ratios_array, axis=1)
    if options["score_samples"]:
        output_names_score_samples = [outputs[2]]
    else:
        output_names_score_samples = None
    score_samples = OnnxReduceMeanApi18(lrd_ratios_array, axes=[1], op_version=opv)
    score_samples.set_onnx_name_prefix("score_samples")
    score_samples_neg = OnnxMul(
        score_samples,
        np.array([-1], dtype=dtype),
        op_version=opv,
        output_names=output_names_score_samples,
    )
    final = OnnxAdd(
        score_samples_neg,
        np.array([-op.offset_], dtype=dtype),
        op_version=opv,
        output_names=[outputs[1]],
    )

    # labels
    # is_inlier = np.ones(X.shape[0], dtype=int)
    # is_inlier[self.decision_function(X) < 0] = -1

    predict = OnnxAdd(
        OnnxMul(
            OnnxCast(
                OnnxLess(final, np.array([0], dtype=dtype), op_version=opv),
                to=TensorProto.INT64,
                op_version=opv,
            ),
            np.array([-2], dtype=np.int64),
            op_version=opv,
        ),
        np.array([1], dtype=np.int64),
        op_version=opv,
        output_names=outputs[0].full_name,
    )
    predict.set_onnx_name_prefix("predict")

    predict.add_to(scope, container)
    final.add_to(scope, container)
    if options["score_samples"]:
        score_samples_neg.add_to(scope, container)


register_converter(
    "SklearnLocalOutlierFactor",
    convert_sklearn_local_outlier_factor,
    options={"score_samples": [True, False], "optim": [None, "cdist"]},
)



# SPDX-License-Identifier: Apache-2.0

import math
import numpy as np
from onnx.numpy_helper import from_array
from sklearn.gaussian_process.kernels import (
    Sum,
    Product,
    ConstantKernel,
    RBF,
    DotProduct,
    ExpSineSquared,
    RationalQuadratic,
    PairwiseKernel,
    WhiteKernel,
)
from ..algebra.complex_functions import onnx_squareform_pdist, onnx_cdist
from ..algebra.onnx_ops import (
    OnnxMul,
    OnnxMatMul,
    OnnxAdd,
    OnnxTranspose,
    OnnxDiv,
    OnnxExp,
    OnnxShape,
    OnnxSin,
    OnnxPow,
    OnnxReduceSumApi11,
    OnnxSqueezeApi11,
    OnnxIdentity,
    OnnxReduceSumSquareApi18,
    OnnxReduceL2_typed,
    OnnxEyeLike,
)
from ..algebra.custom_ops import OnnxCDist

try:
    from ..algebra.onnx_ops import OnnxConstantOfShape
except ImportError:
    OnnxConstantOfShape = None


def convert_kernel_diag(
    kernel, X, output_names=None, dtype=None, optim=None, op_version=None
):
    if op_version is None:
        raise RuntimeError("op_version must not be None.")
    if isinstance(kernel, Sum):
        return OnnxAdd(
            convert_kernel_diag(
                kernel.k1, X, dtype=dtype, optim=optim, op_version=op_version
            ),
            convert_kernel_diag(
                kernel.k2, X, dtype=dtype, optim=optim, op_version=op_version
            ),
            output_names=output_names,
            op_version=op_version,
        )

    if isinstance(kernel, Product):
        return OnnxMul(
            convert_kernel_diag(
                kernel.k1, X, dtype=dtype, optim=optim, op_version=op_version
            ),
            convert_kernel_diag(
                kernel.k2, X, dtype=dtype, optim=optim, op_version=op_version
            ),
            output_names=output_names,
            op_version=op_version,
        )

    if isinstance(kernel, ConstantKernel):
        onnx_zeros = _zero_vector_of_size(
            X, keepdims=0, dtype=dtype, op_version=op_version
        )
        return OnnxAdd(
            onnx_zeros,
            np.array([kernel.constant_value], dtype=dtype),
            output_names=output_names,
            op_version=op_version,
        )

    if isinstance(kernel, (RBF, ExpSineSquared, RationalQuadratic)):
        onnx_zeros = _zero_vector_of_size(
            X, keepdims=0, dtype=dtype, op_version=op_version
        )
        if isinstance(kernel, RBF):
            return OnnxAdd(
                onnx_zeros,
                np.array([1], dtype=dtype),
                output_names=output_names,
                op_version=op_version,
            )
        else:
            return OnnxAdd(
                onnx_zeros,
                np.array([1], dtype=dtype),
                output_names=output_names,
                op_version=op_version,
            )

    if isinstance(kernel, DotProduct):
        t_sigma_0 = py_make_float_array(kernel.sigma_0**2, dtype=dtype)
        return OnnxSqueezeApi11(
            OnnxAdd(
                OnnxReduceSumSquareApi18(X, axes=[1], op_version=op_version),
                t_sigma_0,
                op_version=op_version,
            ),
            output_names=output_names,
            axes=[1],
            op_version=op_version,
        )

    raise RuntimeError(
        "Unable to convert diag method for " "class {}.".format(type(kernel))
    )


def py_make_float_array(cst, dtype, as_tensor=False):
    if dtype not in (np.float32, np.float64):
        raise TypeError(
            "A float array must be of dtype "
            "np.float32 or np.float64 ({}).".format(dtype)
        )
    if isinstance(cst, np.ndarray):
        res = cst.astype(dtype)
    elif not isinstance(cst, (int, float, np.float32, np.float64, np.int32, np.int64)):
        raise TypeError("cst must be a number not {}".format(type(cst)))
    else:
        res = np.array([cst], dtype=dtype)
    return from_array(res) if as_tensor else res


def _convert_exp_sine_squared(
    X,
    Y,
    length_scale=1.2,
    periodicity=1.1,
    pi=math.pi,
    dtype=None,
    optim=None,
    op_version=None,
    **kwargs
):
    """
    Implements the kernel ExpSineSquared.
    """
    if optim is None:
        dists = onnx_cdist(X, Y, metric="euclidean", dtype=dtype, op_version=op_version)
    elif optim == "cdist":
        dists = OnnxCDist(X, Y, metric="euclidean", op_version=op_version)
    else:
        raise ValueError("Unknown optimization '{}'.".format(optim))
    t_pi = py_make_float_array(pi, dtype=dtype)
    t_periodicity = py_make_float_array(periodicity, dtype)
    arg = OnnxMul(
        OnnxDiv(dists, t_periodicity, op_version=op_version),
        t_pi,
        op_version=op_version,
    )
    sin_of_arg = OnnxSin(arg, op_version=op_version)
    t_2 = py_make_float_array(2, dtype=dtype)
    t__2 = py_make_float_array(-2, dtype=dtype)
    t_length_scale = py_make_float_array(length_scale, dtype=dtype)
    K = OnnxExp(
        OnnxMul(
            OnnxPow(
                OnnxDiv(sin_of_arg, t_length_scale, op_version=op_version),
                t_2,
                op_version=op_version,
            ),
            t__2,
            op_version=op_version,
        ),
        op_version=op_version,
    )
    return OnnxIdentity(K, op_version=op_version, **kwargs)


def _convert_dot_product(X, Y, sigma_0=2.0, dtype=None, op_version=None, **kwargs):
    """
    Implements the kernel DotProduct.
    :math:`k(x_i,x_j)=\\sigma_0^2+x_i\\cdot x_j`.
    """
    # It only works in two dimensions.
    t_sigma_0 = py_make_float_array(sigma_0**2, dtype=dtype)
    if isinstance(Y, np.ndarray):
        tr = Y.T.astype(dtype)
    else:
        tr = OnnxTranspose(Y, perm=[1, 0], op_version=op_version)
    matm = OnnxMatMul(X, tr, op_version=op_version)
    K = OnnxAdd(matm, t_sigma_0, op_version=op_version)
    return OnnxIdentity(K, op_version=op_version, **kwargs)


def _convert_rational_quadratic(
    X, Y, length_scale=1.0, alpha=2.0, dtype=None, optim=None, op_version=None, **kwargs
):
    """
    Implements the kernel RationalQuadratic.
    :math:`k(x_i,x_j)=(1 + d(x_i, x_j)^2 / (2*\\alpha * l^2))^{-\\alpha}`.
    """
    if optim is None:
        dists = onnx_cdist(
            X, Y, dtype=dtype, metric="sqeuclidean", op_version=op_version
        )
    elif optim == "cdist":
        dists = OnnxCDist(X, Y, metric="sqeuclidean", op_version=op_version)
    else:
        raise ValueError("Unknown optimization '{}'.".format(optim))
    cst = length_scale**2 * alpha * 2
    t_cst = py_make_float_array(cst, dtype=dtype)
    tmp = OnnxDiv(dists, t_cst, op_version=op_version)
    t_one = py_make_float_array(1, dtype=dtype)
    base = OnnxAdd(tmp, t_one, op_version=op_version)
    t_alpha = py_make_float_array(-alpha, dtype=dtype)
    K = OnnxPow(base, t_alpha, op_version=op_version)
    return OnnxIdentity(K, op_version=op_version, **kwargs)


def _convert_pairwise_kernel(
    X,
    Y,
    metric=None,
    dtype=None,
    op_version=None,
    gamma=None,
    gamma_bounds=None,
    degree=None,
    coef0=None,
    optim=None,
    **kwargs
):
    """
    Implements the kernel PairwiseKernel.

    * cosine: :math:`k(x_i,x_j)=\\frac{<x_i, x_j>}
        {\\parallel x_i \\parallel \\parallel x_j \\parallel}`

    See `KERNEL_PARAMS
    <https://github.com/scikit-learn/scikit-learn/blob/
    95119c13af77c76e150b753485c662b7c52a41a2/sklearn/
    metrics/pairwise.py#L1862>`_.
    """
    if metric == "cosine":
        if isinstance(Y, np.ndarray):
            ny = np.sqrt(np.sum(Y**2, axis=1, keepdims=True))
            norm_y = Y / ny
            norm_try = norm_y.T.astype(dtype)
        else:
            ny = OnnxReduceL2_typed(dtype, Y, axes=[1], op_version=op_version)
            norm_y = OnnxDiv(Y, ny, op_version=op_version)
            norm_try = OnnxTranspose(norm_y, perm=[1, 0], op_version=op_version)

        nx = OnnxReduceL2_typed(dtype, X, axes=[1], op_version=op_version)
        norm_x = OnnxDiv(X, nx, op_version=op_version)
        K = OnnxMatMul(norm_x, norm_try, op_version=op_version)
        return OnnxIdentity(K, op_version=op_version, **kwargs)
    raise NotImplementedError("Metric %r is not implemented." % metric)


def convert_kernel(
    kernel, X, output_names=None, x_train=None, dtype=None, optim=None, op_version=None
):
    if op_version is None:
        raise RuntimeError("op_version must not be None.")
    if isinstance(kernel, Sum):
        clop = OnnxAdd
    elif isinstance(kernel, Product):
        clop = OnnxMul
    else:
        clop = None
    if clop is not None:
        return clop(
            convert_kernel(
                kernel.k1,
                X,
                x_train=x_train,
                dtype=dtype,
                optim=optim,
                op_version=op_version,
            ),
            convert_kernel(
                kernel.k2,
                X,
                x_train=x_train,
                dtype=dtype,
                optim=optim,
                op_version=op_version,
            ),
            output_names=output_names,
            op_version=op_version,
        )

    if isinstance(kernel, ConstantKernel):
        # X and x_train should have the same number of features.
        onnx_zeros_x = _zero_vector_of_size(
            X, keepdims=1, dtype=dtype, op_version=op_version
        )
        if x_train is None:
            onnx_zeros_y = onnx_zeros_x
        else:
            onnx_zeros_y = _zero_vector_of_size(
                x_train, keepdims=1, dtype=dtype, op_version=op_version
            )

        tr = OnnxTranspose(onnx_zeros_y, perm=[1, 0], op_version=op_version)
        mat = OnnxMatMul(onnx_zeros_x, tr, op_version=op_version)
        return OnnxAdd(
            mat,
            np.array([kernel.constant_value], dtype=dtype),
            output_names=output_names,
            op_version=op_version,
        )

    if isinstance(kernel, RBF):
        # length_scale = np.squeeze(length_scale).astype(float)
        zeroh = _zero_vector_of_size(
            X, axis=1, keepdims=0, dtype=dtype, op_version=op_version
        )
        zerov = _zero_vector_of_size(
            X, axis=0, keepdims=1, dtype=dtype, op_version=op_version
        )

        if isinstance(kernel.length_scale, np.ndarray) and len(kernel.length_scale) > 0:
            const = kernel.length_scale.astype(dtype)
        else:
            tensor_value = py_make_float_array(
                kernel.length_scale, dtype=dtype, as_tensor=True
            )
            const = OnnxConstantOfShape(
                OnnxShape(zeroh, op_version=op_version),
                value=tensor_value,
                op_version=op_version,
            )
        X_scaled = OnnxDiv(X, const, op_version=op_version)
        if x_train is None:
            dist = onnx_squareform_pdist(
                X_scaled, metric="sqeuclidean", dtype=dtype, op_version=op_version
            )
        else:
            x_train_scaled = OnnxDiv(x_train, const, op_version=op_version)
            if optim is None:
                dist = onnx_cdist(
                    X_scaled,
                    x_train_scaled,
                    metric="sqeuclidean",
                    dtype=dtype,
                    op_version=op_version,
                )
            elif optim == "cdist":
                dist = OnnxCDist(
                    X_scaled,
                    x_train_scaled,
                    metric="sqeuclidean",
                    op_version=op_version,
                )
            else:
                raise ValueError("Unknown optimization '{}'.".format(optim))

        tensor_value = py_make_float_array(-0.5, dtype=dtype, as_tensor=True)
        cst5 = OnnxConstantOfShape(
            OnnxShape(zerov, op_version=op_version),
            value=tensor_value,
            op_version=op_version,
        )

        # K = np.exp(-.5 * dists)
        exp = OnnxExp(
            OnnxMul(dist, cst5, op_version=op_version),
            output_names=output_names,
            op_version=op_version,
        )

        # This should not be needed.
        # K = squareform(K)
        # np.fill_diagonal(K, 1)
        return exp

    if isinstance(kernel, ExpSineSquared):
        if not isinstance(kernel.length_scale, (float, int)):
            raise NotImplementedError(
                "length_scale should be float not {}.".format(type(kernel.length_scale))
            )

        return _convert_exp_sine_squared(
            X,
            Y=X if x_train is None else x_train,
            length_scale=kernel.length_scale,
            periodicity=kernel.periodicity,
            dtype=dtype,
            output_names=output_names,
            optim=optim,
            op_version=op_version,
        )

    if isinstance(kernel, DotProduct):
        if not isinstance(kernel.sigma_0, (float, int)):
            raise NotImplementedError(
                "sigma_0 should be float not {}.".format(type(kernel.sigma_0))
            )

        if x_train is None:
            return _convert_dot_product(
                X,
                X,
                sigma_0=kernel.sigma_0,
                dtype=dtype,
                output_names=output_names,
                op_version=op_version,
            )
        else:
            if len(x_train.shape) != 2:
                raise NotImplementedError(
                    "Only DotProduct for two dimension train set is " "implemented."
                )
            return _convert_dot_product(
                X,
                x_train,
                sigma_0=kernel.sigma_0,
                dtype=dtype,
                output_names=output_names,
                op_version=op_version,
            )

    if isinstance(kernel, RationalQuadratic):
        if x_train is None:
            return _convert_rational_quadratic(
                X,
                X,
                length_scale=kernel.length_scale,
                dtype=dtype,
                alpha=kernel.alpha,
                output_names=output_names,
                optim=optim,
                op_version=op_version,
            )
        else:
            return _convert_rational_quadratic(
                X,
                x_train,
                length_scale=kernel.length_scale,
                dtype=dtype,
                alpha=kernel.alpha,
                output_names=output_names,
                optim=optim,
                op_version=op_version,
            )

    if isinstance(kernel, PairwiseKernel):
        if x_train is None:
            return _convert_pairwise_kernel(
                X,
                X,
                metric=kernel.metric,
                dtype=dtype,
                output_names=output_names,
                optim=optim,
                op_version=op_version,
            )
        else:
            return _convert_pairwise_kernel(
                X,
                x_train,
                metric=kernel.metric,
                dtype=dtype,
                output_names=output_names,
                optim=optim,
                op_version=op_version,
            )

    if isinstance(kernel, WhiteKernel):
        # X and x_train should have the same number of features.
        onnx_zeros_x = _zero_vector_of_size(
            X, keepdims=1, dtype=dtype, op_version=op_version
        )
        if x_train is None:
            onnx_zeros_y = onnx_zeros_x
        else:
            onnx_zeros_y = _zero_vector_of_size(
                x_train, keepdims=1, dtype=dtype, op_version=op_version
            )
        tr = OnnxTranspose(onnx_zeros_y, perm=[1, 0], op_version=op_version)
        mat = OnnxMatMul(onnx_zeros_x, tr, op_version=op_version)

        if x_train is not None:
            return OnnxIdentity(mat, op_version=op_version, output_names=output_names)

        return OnnxMul(
            OnnxEyeLike(mat, op_version=op_version),
            OnnxIdentity(
                np.array([kernel.noise_level], dtype=dtype), op_version=op_version
            ),
            op_version=op_version,
            output_names=output_names,
        )

    raise RuntimeError(
        "Unable to convert __call__ method for " "class {}.".format(type(kernel))
    )


def _zero_vector_of_size(
    X, output_names=None, axis=0, keepdims=None, dtype=None, op_version=None
):
    if op_version is None:
        raise RuntimeError("op_version must not be None.")
    if keepdims is None:
        raise ValueError("Default for keepdims is not allowed.")
    if dtype == np.float32:
        res = OnnxReduceSumApi11(
            OnnxConstantOfShape(
                OnnxShape(X, op_version=op_version), op_version=op_version
            ),
            axes=[1 - axis],
            keepdims=keepdims,
            output_names=output_names,
            op_version=op_version,
        )
    elif dtype in (np.float64, np.int32, np.int64):
        res = OnnxReduceSumApi11(
            OnnxConstantOfShape(
                OnnxShape(X, op_version=op_version),
                value=py_make_float_array(0, dtype=dtype, as_tensor=True),
                op_version=op_version,
            ),
            axes=[1 - axis],
            keepdims=keepdims,
            output_names=output_names,
            op_version=op_version,
        )
    else:
        raise NotImplementedError(
            "Unable to create zero vector of type {}".format(dtype)
        )
    return res



# SPDX-License-Identifier: Apache-2.0

import logging
from ..common._registration import register_shape_calculator, get_shape_calculator
from .._supported_operators import sklearn_operator_name_map


def convert_sklearn_grid_search_cv(operator):
    grid_search_op = operator.raw_operator
    best_estimator = grid_search_op.best_estimator_
    name = sklearn_operator_name_map.get(type(best_estimator), None)
    if name is None:
        logger = logging.getLogger("skl2onnx")
        logger.warn(
            "[convert_sklearn_grid_search_cv] failed to find alias "
            "to model type %r.",
            type(best_estimator),
        )
        return
    op = operator.new_raw_operator(best_estimator, name)
    shape_calc = get_shape_calculator(name)
    shape_calc(op)
    operator.outputs = op.outputs


register_shape_calculator("SklearnGridSearchCV", convert_sklearn_grid_search_cv)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._registration import register_shape_calculator
from ..common.data_types import Int64TensorType, FloatTensorType


def calculate_sklearn_ordinal_encoder_output_shapes(operator):
    ordinal_op = operator.raw_operator
    op_features = sum(list(map(lambda x: x.type.shape[1], operator.inputs)))
    if np.issubdtype(ordinal_op.dtype, np.floating):
        operator.outputs[0].type = FloatTensorType(
            [operator.inputs[0].get_first_dimension(), op_features]
        )
    else:
        operator.outputs[0].type = Int64TensorType(
            [operator.inputs[0].get_first_dimension(), op_features]
        )


register_shape_calculator(
    "SklearnOrdinalEncoder", calculate_sklearn_ordinal_encoder_output_shapes
)



# SPDX-License-Identifier: Apache-2.0


from inspect import signature
from collections import OrderedDict


def check_signature(fct, reference, skip=None):
    """
    Checks that two functions have the same signature
    (same parameter names).
    Raises an exception otherwise.
    """

    def select_parameters(pars):
        new_pars = OrderedDict()
        for i, (name, p) in enumerate(pars.items()):
            if i >= 3 and name in ("op_type", "op_domain", "op_version"):
                if p.default is not None:
                    # Parameters op_type and op_domain are skipped.
                    continue
            new_pars[name] = p
        return new_pars

    sig = signature(fct)
    sig_ref = signature(reference)
    fct_pars = select_parameters(sig.parameters)
    ref_pars = select_parameters(sig_ref.parameters)
    if len(fct_pars) != len(ref_pars):
        raise TypeError(
            "Function '{}' must have {} parameters but has {}."
            "".format(fct.__name__, len(ref_pars), len(fct_pars))
        )
    for i, (a, b) in enumerate(zip(fct_pars, ref_pars)):
        if a != b and skip is not None and b not in skip and a not in skip:
            raise NameError(
                "Parameter name mismatch at position {}."
                "Function '{}' has '{}' but '{}' is expected."
                "".format(i + 1, fct.__name__, a, b)
            )



# SPDX-License-Identifier: Apache-2.0

"""
Common functions to convert any learner based on trees.
"""
import numpy as np


def get_default_tree_classifier_attribute_pairs():
    attrs = {}
    attrs["post_transform"] = "NONE"
    attrs["nodes_treeids"] = []
    attrs["nodes_nodeids"] = []
    attrs["nodes_featureids"] = []
    attrs["nodes_modes"] = []
    attrs["nodes_values"] = []
    attrs["nodes_truenodeids"] = []
    attrs["nodes_falsenodeids"] = []
    attrs["nodes_missing_value_tracks_true"] = []
    attrs["nodes_hitrates"] = []
    attrs["class_treeids"] = []
    attrs["class_nodeids"] = []
    attrs["class_ids"] = []
    attrs["class_weights"] = []
    return attrs


def get_default_tree_regressor_attribute_pairs():
    attrs = {}
    attrs["post_transform"] = "NONE"
    attrs["n_targets"] = 0
    attrs["nodes_treeids"] = []
    attrs["nodes_nodeids"] = []
    attrs["nodes_featureids"] = []
    attrs["nodes_modes"] = []
    attrs["nodes_values"] = []
    attrs["nodes_truenodeids"] = []
    attrs["nodes_falsenodeids"] = []
    attrs["nodes_missing_value_tracks_true"] = []
    attrs["nodes_hitrates"] = []
    attrs["target_treeids"] = []
    attrs["target_nodeids"] = []
    attrs["target_ids"] = []
    attrs["target_weights"] = []
    return attrs


def find_switch_point(fy, nfy):
    """
    Finds the double so that
    ``(float)x != (float)(x + espilon)``.
    """
    a = np.float64(fy)
    b = np.float64(nfy)
    fa = np.float32(a)
    a0, b0 = a, a
    while a != a0 or b != b0:
        a0, b0 = a, b
        m = (a + b) / 2
        fm = np.float32(m)
        if fm == fa:
            a = m
            fa = fm
        else:
            b = m
    return a


def sklearn_threshold(dy, dtype, mode):
    """
    *scikit-learn* does not compare x to a threshold
    but (float)x to a double threshold. As we need a float
    threshold, we need a different value than the threshold
    rounded to float. For floats, it finds float *w* which
    verifies::

        (float)x <= y <=> (float)x <= w

    For doubles, it finds double *w* which verifies::

        (float)x <= y <=> x <= w
    """
    if mode == "BRANCH_LEQ":
        if dtype == np.float32:
            fy = np.float32(dy)
            if fy == dy:
                return np.float64(fy)
            if fy < dy:
                return np.float64(fy)
            eps = max(abs(fy), np.finfo(np.float32).eps) * 10
            nfy = np.nextafter([fy], [fy - eps], dtype=np.float32)[0]
            return np.float64(nfy)
        elif dtype == np.float64:
            fy = np.float32(dy)
            eps = max(abs(fy), np.finfo(np.float32).eps) * 10
            afy = np.nextafter([fy], [fy - eps], dtype=np.float32)[0]
            afy2 = find_switch_point(afy, fy)
            if fy > dy > afy2:
                return afy2
            bfy = np.nextafter([fy], [fy + eps], dtype=np.float32)[0]
            bfy2 = find_switch_point(fy, bfy)
            if fy <= dy <= bfy2:
                return bfy2
            return np.float64(fy)
        raise TypeError("Unexpected dtype {}.".format(dtype))
    raise RuntimeError(
        "Threshold is not changed for other mode and "
        "'BRANCH_LEQ' (actually '{}').".format(mode)
    )


def add_node(
    attr_pairs,
    is_classifier,
    tree_id,
    tree_weight,
    node_id,
    feature_id,
    mode,
    value,
    true_child_id,
    false_child_id,
    weights,
    weight_id_bias,
    leaf_weights_are_counts,
    adjust_threshold_for_sklearn,
    dtype,
    nodes_missing_value_tracks_true=False,
):
    attr_pairs["nodes_treeids"].append(tree_id)
    attr_pairs["nodes_nodeids"].append(node_id)
    attr_pairs["nodes_featureids"].append(feature_id)
    attr_pairs["nodes_modes"].append(mode)
    if adjust_threshold_for_sklearn and mode != "LEAF":
        attr_pairs["nodes_values"].append(sklearn_threshold(value, dtype, mode))
    else:
        attr_pairs["nodes_values"].append(value)
    attr_pairs["nodes_truenodeids"].append(true_child_id)
    attr_pairs["nodes_falsenodeids"].append(false_child_id)
    attr_pairs["nodes_missing_value_tracks_true"].append(
        nodes_missing_value_tracks_true
    )
    attr_pairs["nodes_hitrates"].append(1.0)

    # Add leaf information for making prediction
    if mode == "LEAF":
        flattened_weights = weights.flatten()
        factor = tree_weight
        # If the values stored at leaves are counts of possible classes, we
        # need convert them to probabilities by doing a normalization.
        if leaf_weights_are_counts:
            s = sum(flattened_weights)
            factor /= float(s) if s != 0.0 else 1.0
        flattened_weights = [w * factor for w in flattened_weights]
        if len(flattened_weights) == 2 and is_classifier:
            flattened_weights = [flattened_weights[1]]

        # Note that attribute names for making prediction are different for
        # classifiers and regressors
        if is_classifier:
            for i, w in enumerate(flattened_weights):
                attr_pairs["class_treeids"].append(tree_id)
                attr_pairs["class_nodeids"].append(node_id)
                attr_pairs["class_ids"].append(i + weight_id_bias)
                attr_pairs["class_weights"].append(w)
        else:
            for i, w in enumerate(flattened_weights):
                attr_pairs["target_treeids"].append(tree_id)
                attr_pairs["target_nodeids"].append(node_id)
                attr_pairs["target_ids"].append(i + weight_id_bias)
                attr_pairs["target_weights"].append(w)


def add_tree_to_attribute_pairs(
    attr_pairs,
    is_classifier,
    tree,
    tree_id,
    tree_weight,
    weight_id_bias,
    leaf_weights_are_counts,
    adjust_threshold_for_sklearn=False,
    dtype=None,
):
    for i in range(tree.node_count):
        node_id = i
        weight = tree.value[i]

        if tree.children_left[i] > i or tree.children_right[i] > i:
            mode = "BRANCH_LEQ"
            feat_id = tree.feature[i]
            threshold = tree.threshold[i]
            left_child_id = int(tree.children_left[i])
            right_child_id = int(tree.children_right[i])
        else:
            mode = "LEAF"
            feat_id = 0
            threshold = 0.0
            left_child_id = 0
            right_child_id = 0

        add_node(
            attr_pairs,
            is_classifier,
            tree_id,
            tree_weight,
            node_id,
            feat_id,
            mode,
            threshold,
            left_child_id,
            right_child_id,
            weight,
            weight_id_bias,
            leaf_weights_are_counts,
            adjust_threshold_for_sklearn=adjust_threshold_for_sklearn,
            dtype=dtype,
        )


def add_tree_to_attribute_pairs_hist_gradient_boosting(
    attr_pairs,
    is_classifier,
    tree,
    tree_id,
    tree_weight,
    weight_id_bias,
    leaf_weights_are_counts,
    adjust_threshold_for_sklearn=False,
    dtype=None,
):
    for i, node in enumerate(tree.nodes):
        node_id = i
        weight = node["value"]

        if node["is_leaf"]:
            mode = "LEAF"
            feat_id = 0
            threshold = 0.0
            left_child_id = 0
            right_child_id = 0
            missing = False
        else:
            mode = "BRANCH_LEQ"
            feat_id = node["feature_idx"]
            try:
                threshold = node["threshold"]
            except ValueError:
                threshold = node["num_threshold"]
            left_child_id = node["left"]
            right_child_id = node["right"]
            missing = node["missing_go_to_left"]

        add_node(
            attr_pairs,
            is_classifier,
            tree_id,
            tree_weight,
            node_id,
            feat_id,
            mode,
            threshold,
            left_child_id,
            right_child_id,
            weight,
            weight_id_bias,
            leaf_weights_are_counts,
            adjust_threshold_for_sklearn=adjust_threshold_for_sklearn,
            dtype=dtype,
            nodes_missing_value_tracks_true=missing,
        )



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..algebra.onnx_ops import OnnxConcat, OnnxReshapeApi13, OnnxIdentity

try:
    from ..algebra.onnx_ops import OnnxSequenceConstruct
except ImportError:
    # Available since opset 11
    OnnxSequenceConstruct = None
from ..algebra.onnx_operator import OnnxSubEstimator


def convert_multi_output_regressor_converter(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converts a *MultiOutputRegressor* into *ONNX* format.
    """
    op_version = container.target_opset
    op = operator.raw_operator
    inp = operator.inputs[0]
    y_list = [
        OnnxReshapeApi13(
            OnnxSubEstimator(sub, inp, op_version=op_version),
            np.array([-1, 1], dtype=np.int64),
            op_version=op_version,
        )
        for sub in op.estimators_
    ]

    output = OnnxConcat(
        *y_list, axis=1, op_version=op_version, output_names=[operator.outputs[0]]
    )
    output.add_to(scope=scope, container=container)


def convert_multi_output_classifier_converter(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converts a *MultiOutputClassifier* into *ONNX* format.
    """
    if OnnxSequenceConstruct is None:
        raise RuntimeError("This converter requires opset>=11.")
    op_version = container.target_opset
    op_version = container.target_opset
    op = operator.raw_operator
    inp = operator.inputs[0]
    options = scope.get_options(op)
    if options.get("nocl", True):
        options = options.copy()
    else:
        options = {}

    options.update({"zipmap": False})
    y_list = [
        OnnxSubEstimator(sub, inp, op_version=op_version, options=options)
        for sub in op.estimators_
    ]

    # labels
    label_list = [
        OnnxReshapeApi13(y[0], np.array([-1, 1], dtype=np.int64), op_version=op_version)
        for y in y_list
    ]

    # probabilities
    proba_list = [OnnxIdentity(y[1], op_version=op_version) for y in y_list]
    label = OnnxConcat(
        *label_list, axis=1, op_version=op_version, output_names=[operator.outputs[0]]
    )
    label.add_to(scope=scope, container=container)

    proba = OnnxSequenceConstruct(
        *proba_list, op_version=op_version, output_names=[operator.outputs[1]]
    )
    proba.add_to(scope=scope, container=container)


register_converter(
    "SklearnMultiOutputRegressor", convert_multi_output_regressor_converter
)
register_converter(
    "SklearnMultiOutputClassifier",
    convert_multi_output_classifier_converter,
    options={
        "nocl": [False, True],
        "output_class_labels": [False, True],
        "zipmap": [False, True],
    },
)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import Int64TensorType, guess_numpy_type, guess_proto_type
from ..algebra.onnx_ops import OnnxAdd, OnnxCast, OnnxDiv, OnnxMatMul, OnnxSub


def convert_pls_regression(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    X = operator.inputs[0]
    op = operator.raw_operator
    opv = container.target_opset
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    if isinstance(X.type, Int64TensorType):
        X = OnnxCast(X, to=proto_dtype, op_version=opv)

    coefs = op.x_mean_ if hasattr(op, "x_mean_") else op._x_mean
    std = op.x_std_ if hasattr(op, "x_std_") else op._x_std
    ym = op.y_mean_ if hasattr(op, "x_mean_") else op._y_mean

    norm_x = OnnxDiv(
        OnnxSub(X, coefs.astype(dtype), op_version=opv),
        std.astype(dtype),
        op_version=opv,
    )
    if hasattr(op, "set_predict_request"):
        # new in 1.3
        coefs = op.coef_.T.astype(dtype)
    else:
        coefs = op.coef_.astype(dtype)
    dot = OnnxMatMul(norm_x, coefs, op_version=opv)
    pred = OnnxAdd(dot, ym.astype(dtype), op_version=opv, output_names=operator.outputs)
    pred.add_to(scope, container)


register_converter("SklearnPLSRegression", convert_pls_regression)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_replace_transformer(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    operator.outputs[0].type = operator.inputs[0].type


register_shape_calculator(
    "SklearnReplaceTransformer", calculate_sklearn_replace_transformer
)



# SPDX-License-Identifier: Apache-2.0

import pprint
import numpy as np
from onnx import TensorProto
from ..common._apply_operation import (
    apply_abs,
    apply_add,
    apply_cast,
    apply_concat,
    apply_clip,
    apply_div,
    apply_exp,
    apply_mul,
    apply_reshape,
    apply_sub,
)
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_numpy_type, Int64TensorType, guess_proto_type
from ..common._registration import register_converter
from .._supported_operators import sklearn_operator_name_map
from sklearn.ensemble import RandomForestClassifier


def _handle_zeros(
    scope, container, concatenated_prob_name, reduced_prob_name, n_classes, proto_type
):
    """
    This function replaces 0s in concatenated_prob_name with 1s and
    0s in reduced_prob_name with n_classes.
    """
    cast_prob_name = scope.get_unique_variable_name("cast_prob")
    bool_not_cast_prob_name = scope.get_unique_variable_name("bool_not_cast_prob")
    mask_name = scope.get_unique_variable_name("mask")
    masked_concatenated_prob_name = scope.get_unique_variable_name(
        "masked_concatenated_prob"
    )
    n_classes_name = scope.get_unique_variable_name("n_classes")
    reduced_prob_mask_name = scope.get_unique_variable_name("reduced_prob_mask")
    masked_reduced_prob_name = scope.get_unique_variable_name("masked_reduced_prob")

    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT

    container.add_initializer(n_classes_name, proto_type2, [], [n_classes])

    apply_cast(scope, reduced_prob_name, cast_prob_name, container, to=TensorProto.BOOL)
    container.add_node(
        "Not",
        cast_prob_name,
        bool_not_cast_prob_name,
        name=scope.get_unique_operator_name("Not"),
    )
    apply_cast(scope, bool_not_cast_prob_name, mask_name, container, to=proto_type2)
    apply_add(
        scope,
        [concatenated_prob_name, mask_name],
        masked_concatenated_prob_name,
        container,
        broadcast=1,
    )
    apply_mul(
        scope,
        [mask_name, n_classes_name],
        reduced_prob_mask_name,
        container,
        broadcast=1,
    )
    apply_add(
        scope,
        [reduced_prob_name, reduced_prob_mask_name],
        masked_reduced_prob_name,
        container,
        broadcast=0,
    )
    return masked_concatenated_prob_name, masked_reduced_prob_name


def _transform_sigmoid(scope, container, model, df_col_name, k, proto_type):
    """
    Sigmoid Calibration method
    """
    a_name = scope.get_unique_variable_name("a")
    b_name = scope.get_unique_variable_name("b")
    a_df_prod_name = scope.get_unique_variable_name("a_df_prod")
    exp_parameter_name = scope.get_unique_variable_name("exp_parameter")
    exp_result_name = scope.get_unique_variable_name("exp_result")
    unity_name = scope.get_unique_variable_name("unity")
    denominator_name = scope.get_unique_variable_name("denominator")
    sigmoid_predict_result_name = scope.get_unique_variable_name(
        "sigmoid_predict_result"
    )

    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT

    if hasattr(model, "calibrators_"):
        # scikit-learn<1.1
        calibrators = model.calibrators_
    elif hasattr(model, "calibrators"):
        # scikit-learn>=1.1
        calibrators = model.calibrators
    else:
        raise AttributeError(
            "Unable to find attribute calibrators_ or "
            "calibrators, check the model was trained, "
            "type=%r." % type(model)
        )

    container.add_initializer(a_name, proto_type2, [], [calibrators[k].a_])
    container.add_initializer(b_name, proto_type2, [], [calibrators[k].b_])
    container.add_initializer(unity_name, proto_type2, [], [1])

    apply_mul(scope, [a_name, df_col_name], a_df_prod_name, container, broadcast=0)
    apply_add(
        scope, [a_df_prod_name, b_name], exp_parameter_name, container, broadcast=0
    )
    apply_exp(scope, exp_parameter_name, exp_result_name, container)
    apply_add(
        scope, [unity_name, exp_result_name], denominator_name, container, broadcast=0
    )
    apply_div(
        scope,
        [unity_name, denominator_name],
        sigmoid_predict_result_name,
        container,
        broadcast=0,
    )
    return sigmoid_predict_result_name


def _transform_isotonic(scope, container, model, T, k, dtype, proto_type):
    """
    Isotonic calibration method
    This function can only handle one instance at a time because
    ArrayFeatureExtractor can only extract based on the last axis,
    so we can't fetch different columns for different rows.
    """
    if hasattr(model, "calibrators_"):
        # scikit-learn<1.1
        calibrators = model.calibrators_
    elif hasattr(model, "calibrators"):
        # scikit-learn>=1.1
        calibrators = model.calibrators
    else:
        raise AttributeError(
            "Unable to find attribute calibrators_ or "
            "calibrators, check the model was trained, "
            "type=%r." % type(model)
        )

    if calibrators[k].out_of_bounds == "clip":
        clipped_df_name = scope.get_unique_variable_name("clipped_df")
        apply_clip(
            scope,
            T,
            clipped_df_name,
            container,
            operator_name=scope.get_unique_operator_name("Clip"),
            max=np.array(calibrators[k].X_max_, dtype=dtype),
            min=np.array(calibrators[k].X_min_, dtype=dtype),
        )
        T = clipped_df_name

    reshaped_df_name = scope.get_unique_variable_name("reshaped_df")
    calibrator_x_name = scope.get_unique_variable_name("calibrator_x")
    calibrator_y_name = scope.get_unique_variable_name("calibrator_y")
    distance_name = scope.get_unique_variable_name("distance")
    absolute_distance_name = scope.get_unique_variable_name("absolute_distance")
    nearest_x_index_name = scope.get_unique_variable_name("nearest_x_index")
    nearest_y_name = scope.get_unique_variable_name("nearest_y")

    if hasattr(calibrators[k], "_X_"):
        atX, atY = "_X_", "_y_"
    elif hasattr(calibrators[k], "_necessary_X_"):
        atX, atY = "_necessary_X_", "_necessary_y_"
    elif hasattr(calibrators[k], "X_thresholds_"):
        atX, atY = "X_thresholds_", "y_thresholds_"
    else:
        raise AttributeError(
            "Unable to find attribute '_X_' or '_necessary_X_' "
            "for type {}\n{}."
            "".format(type(calibrators[k]), pprint.pformat(dir(calibrators[k])))
        )

    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT

    container.add_initializer(
        calibrator_x_name,
        proto_type2,
        [len(getattr(calibrators[k], atX))],
        getattr(calibrators[k], atX),
    )
    container.add_initializer(
        calibrator_y_name,
        proto_type2,
        [len(getattr(calibrators[k], atY))],
        getattr(calibrators[k], atY),
    )

    apply_reshape(scope, T, reshaped_df_name, container, desired_shape=(-1, 1))
    apply_sub(
        scope,
        [reshaped_df_name, calibrator_x_name],
        distance_name,
        container,
        broadcast=1,
    )
    apply_abs(scope, distance_name, absolute_distance_name, container)
    container.add_node(
        "ArgMin",
        absolute_distance_name,
        nearest_x_index_name,
        axis=1,
        name=scope.get_unique_operator_name("ArgMin"),
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [calibrator_y_name, nearest_x_index_name],
        nearest_y_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )

    nearest_y_name_reshaped = scope.get_unique_variable_name("nearest_y_name_reshaped")
    apply_reshape(
        scope, nearest_y_name, nearest_y_name_reshaped, container, desired_shape=(-1, 1)
    )
    return nearest_y_name_reshaped


def convert_calibrated_classifier_base_estimator(
    scope, operator, container, model, model_index
):
    # Computational graph:
    #
    # In the following graph, variable names are in lower case characters only
    # and operator names are in upper case characters. We borrow operator names
    # from the official ONNX spec:
    # https://github.com/onnx/onnx/blob/master/docs/Operators.md
    # All variables are followed by their shape in [].
    #
    # Symbols:
    # M: Number of instances
    # N: Number of features
    # C: Number of classes
    # CLASSIFIERCONVERTER: classifier converter corresponding to the op_type
    # a: slope in sigmoid model
    # b: intercept in sigmoid model
    # k: variable in the range [0, C)
    # input: input
    # class_prob_tensor: tensor with class probabilities(function output)
    #
    # Graph:
    #
    #   input [M, N] -> CLASSIFIERCONVERTER -> label [M]
    #                          |
    #                          V
    #                    probability_tensor [M, C]
    #                          |
    #         .----------------'---------.
    #         |                          |
    #         V                          V
    # ARRAYFEATUREEXTRACTOR <- k [1] -> ARRAYFEATUREEXTRACTOR
    #         |                          |
    #         V                          V
    #  transposed_df_col[M, 1] transposed_df_col[M, 1]
    #       |--------------------------|----------.--------------------------.
    #       |                          |          |                          |
    #       |if model.method='sigmoid' |          |if model.method='isotonic'|
    #       |                          |          |                          |
    #       V                          V          |if out_of_bounds='clip'   |
    #      MUL <-------- a -------->  MUL         V                          V
    #       |                          |          CLIP     ...             CLIP
    #       V                          V          |                          |
    #  a_df_prod [M, 1]  ... a_df_prod [M, 1]     V                          V
    #       |                          |  clipped_df [M, 1]...clipped_df [M, 1]
    #       V                          V          |                          |
    #      ADD <--------- b ---------> ADD        '-------------------.------'
    #         |                          |                            |
    #         V                          V                            |
    #  exp_parameter [M, 1] ...   exp_parameter [M, 1]                |
    #         |                          |                            |
    #         V                          V                            |
    #        EXP        ...             EXP                           |
    #         |                          |                            |
    #         V                          V                            |
    #  exp_result [M, 1]  ...    exp_result [M, 1]                    |
    #         |                          |                            |
    #         V                          V                            |
    #       ADD <------- unity -------> ADD                           |
    #         |                          |                            |
    #         V                          V                            |
    #  denominator [M, 1]  ...   denominator [M, 1]                   |
    #         |                          |                            |
    #         V                          V                            |
    #        DIV <------- unity ------> DIV                           |
    #         |                          |                            |
    #         V                          V                            |
    # sigmoid_predict_result [M, 1] ... sigmoid_predict_result [M, 1] |
    #         |                          |                            |
    #         '-----.--------------------'                            |
    #               |-------------------------------------------------'
    #               |
    #               V
    #            CONCAT -> concatenated_prob [M, C]
    #                          |
    #        if  C = 2         |  if C != 2
    #      .-------------------'---------------------------.---------.
    #      |                                               |         |
    #      V                                               |         V
    # ARRAYFEATUREEXTRACTOR <- col_number [1]              |    REDUCESUM
    #                   |                                  |         |
    #                   '--------------------------------. |         |
    # unit_float_tensor [1] -> SUB <- first_col [M, 1] <-' |         |
    #                           |                         /          |
    #                           V                        V           V
    #                         CONCAT                    DIV <- reduced_prob [M]
    #                           |                        |
    #                           V                        |
    #                        class_prob_tensor [M, C] <--'
    model_proba = {RandomForestClassifier}

    if scope.get_options(operator.raw_operator, dict(nocl=False))["nocl"]:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__
            )
        )
    proto_type = guess_proto_type(operator.inputs[0].type)
    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    base_model = (
        model.estimator if hasattr(model, "estimator") else model.base_estimator
    )
    op_type = sklearn_operator_name_map[type(base_model)]
    n_classes = (
        len(model.classes_) if hasattr(model, "classes_") else len(base_model.classes_)
    )
    prob_name = [None] * n_classes

    this_operator = scope.declare_local_operator(op_type, base_model)
    if (
        container.has_options(base_model, "raw_scores")
        and type(base_model) not in model_proba
    ):
        container.add_options(id(base_model), {"raw_scores": True})
        scope.add_options(id(base_model), {"raw_scores": True})
    this_operator.inputs = operator.inputs
    label_name = scope.declare_local_variable("label", Int64TensorType())
    df_name = scope.declare_local_variable(
        "uncal_probability", operator.inputs[0].type.__class__()
    )
    this_operator.outputs.append(label_name)
    this_operator.outputs.append(df_name)
    df_inp = df_name.full_name

    for k in range(n_classes):
        cur_k = k
        if n_classes == 2:
            cur_k += 1
        k_name = scope.get_unique_variable_name("k")
        df_col_name = scope.get_unique_variable_name(
            "tdf_col_%d_c%d" % (model_index, k)
        )
        prob_name[k] = scope.get_unique_variable_name(
            "prob_{}_c{}".format(model_index, k)
        )

        container.add_initializer(k_name, TensorProto.INT64, [], [cur_k])

        container.add_node(
            "ArrayFeatureExtractor",
            [df_inp, k_name],
            df_col_name,
            name=scope.get_unique_operator_name("CaliAFE_%d_c%d" % (model_index, k)),
            op_domain="ai.onnx.ml",
        )
        if model.method == "sigmoid":
            T = _transform_sigmoid(scope, container, model, df_col_name, k, proto_type)
        else:
            T = _transform_isotonic(
                scope, container, model, df_col_name, k, dtype, proto_type
            )

        prob_name[k] = T
        if n_classes == 2:
            break

    if n_classes == 2:
        zeroth_col_name = scope.get_unique_variable_name("zeroth_col%d" % model_index)
        merged_prob_name = scope.get_unique_variable_name("merged_prob%d" % model_index)
        unit_float_tensor_name = scope.get_unique_variable_name(
            "unit_float_tensor%d" % model_index
        )

        container.add_initializer(unit_float_tensor_name, proto_type2, [], [1.0])

        apply_sub(
            scope,
            [unit_float_tensor_name, prob_name[0]],
            zeroth_col_name,
            container,
            broadcast=1,
        )
        apply_concat(
            scope,
            [zeroth_col_name, prob_name[0]],
            merged_prob_name,
            container,
            axis=1,
            operator_name=scope.get_unique_variable_name("CaliConc%d" % model_index),
        )
        class_prob_tensor_name = merged_prob_name
    else:
        concatenated_prob_name = scope.get_unique_variable_name("concatenated_prob")
        reduced_prob_name = scope.get_unique_variable_name("reduced_prob")
        calc_prob_name = scope.get_unique_variable_name("calc_prob")

        apply_concat(scope, prob_name, concatenated_prob_name, container, axis=1)
        if container.target_opset < 13:
            container.add_node(
                "ReduceSum",
                concatenated_prob_name,
                reduced_prob_name,
                axes=[1],
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        else:
            axis_name = scope.get_unique_variable_name("axis")
            container.add_initializer(axis_name, TensorProto.INT64, [1], [1])
            container.add_node(
                "ReduceSum",
                [concatenated_prob_name, axis_name],
                reduced_prob_name,
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        num, deno = _handle_zeros(
            scope,
            container,
            concatenated_prob_name,
            reduced_prob_name,
            n_classes,
            proto_type,
        )
        apply_div(
            scope,
            [num, deno],
            calc_prob_name,
            container,
            broadcast=1,
            operator_name=scope.get_unique_variable_name("CaliDiv%d" % model_index),
        )
        class_prob_tensor_name = calc_prob_name
    return class_prob_tensor_name


def convert_sklearn_calibrated_classifier_cv(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    # Computational graph:
    #
    # In the following graph, variable names are in lower case characters only
    # and operator names are in upper case characters. We borrow operator names
    # from the official ONNX spec:
    # https://github.com/onnx/onnx/blob/master/docs/Operators.md
    # All variables are followed by their shape in [].
    #
    # Symbols:
    # M: Number of instances
    # N: Number of features
    # C: Number of classes
    # CONVERT_BASE_ESTIMATOR: base estimator convert function defined above
    # clf_length: number of calibrated classifiers
    # input: input
    # output: output
    # class_prob: class probabilities
    #
    # Graph:
    #
    #                         input [M, N]
    #                               |
    #           .-------------------|--------------------------.
    #           |                   |                          |
    #           V                   V                          V
    # CONVERT_BASE_ESTIMATOR  CONVERT_BASE_ESTIMATOR ... CONVERT_BASE_ESTIMATOR
    #           |                   |                          |
    #           V                   V                          V
    #  prob_scores_0 [M, C] prob_scores_1 [M, C] ... prob_scores_(clf_length-1)
    #           |                   |                          |  [M, C]
    #           '-------------------|--------------------------'
    #                               V
    #       add_result [M, C] <--- SUM
    #           |
    #           '--> DIV <- clf_length [1]
    #                 |
    #                 V
    #            class_prob [M, C] -> ARGMAX -> argmax_output [M, 1]
    #                                                   |
    #             classes -> ARRAYFEATUREEXTRACTOR  <---'
    #                               |
    #                               V
    #                            output [1]

    op = operator.raw_operator
    classes = op.classes_
    output_shape = (-1,)
    class_type = TensorProto.STRING
    proto_type = guess_proto_type(operator.inputs[0].type)
    proto_type2 = proto_type
    if proto_type2 not in (TensorProto.FLOAT, TensorProto.DOUBLE):
        proto_type2 = TensorProto.FLOAT

    if np.issubdtype(op.classes_.dtype, np.floating):
        class_type = TensorProto.INT32
        classes = classes.astype(np.int32)
    elif (
        np.issubdtype(op.classes_.dtype, np.signedinteger)
        or op.classes_.dtype == np.bool_
    ):
        class_type = TensorProto.INT32
    else:
        classes = np.array([s.encode("utf-8") for s in classes])

    clf_length = len(op.calibrated_classifiers_)
    prob_scores_name = []

    clf_length_name = scope.get_unique_variable_name("clf_length")
    classes_name = scope.get_unique_variable_name("classes")
    reshaped_result_name = scope.get_unique_variable_name("reshaped_result")
    argmax_output_name = scope.get_unique_variable_name("argmax_output")
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        "array_feature_extractor_result"
    )
    add_result_name = scope.get_unique_variable_name("add_result")

    container.add_initializer(classes_name, class_type, classes.shape, classes)
    container.add_initializer(clf_length_name, proto_type2, [], [clf_length])

    for clf_index, clf in enumerate(op.calibrated_classifiers_):
        prob_scores_name.append(
            convert_calibrated_classifier_base_estimator(
                scope, operator, container, clf, clf_index
            )
        )

    container.add_node(
        "Sum",
        [s for s in prob_scores_name],
        add_result_name,
        op_version=7,
        name=scope.get_unique_operator_name("Sum"),
    )
    apply_div(
        scope,
        [add_result_name, clf_length_name],
        operator.outputs[1].full_name,
        container,
        broadcast=1,
    )
    class_prob_name = operator.outputs[1].full_name
    container.add_node(
        "ArgMax",
        class_prob_name,
        argmax_output_name,
        name=scope.get_unique_operator_name("ArgMax"),
        axis=1,
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [classes_name, argmax_output_name],
        array_feature_extractor_result_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )

    if class_type == TensorProto.INT32:
        apply_reshape(
            scope,
            array_feature_extractor_result_name,
            reshaped_result_name,
            container,
            desired_shape=output_shape,
        )
        apply_cast(
            scope,
            reshaped_result_name,
            operator.outputs[0].full_name,
            container,
            to=TensorProto.INT64,
        )
    else:
        apply_reshape(
            scope,
            array_feature_extractor_result_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=output_shape,
        )


register_converter(
    "SklearnCalibratedClassifierCV",
    convert_sklearn_calibrated_classifier_cv,
    options={
        "zipmap": [True, False, "columns"],
        "output_class_labels": [False, True],
        "nocl": [True, False],
    },
)



# SPDX-License-Identifier: Apache-2.0


import copy
from ..common._registration import register_shape_calculator


def calculate_sklearn_function_transformer_output_shapes(operator):
    """
    This operator is used only to merge columns in a pipeline.
    Only identity function is supported.
    """
    if operator.raw_operator.func is not None:
        raise RuntimeError(
            "FunctionTransformer is not supported unless the "
            "transform function is None (= identity). "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues."
        )
    N = operator.inputs[0].get_first_dimension()
    C = 0
    for variable in operator.inputs:
        if variable.type.shape[1] is not None:
            C += variable.type.shape[1]
        else:
            C = None
            break

    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape = [N, C]


register_shape_calculator(
    "SklearnFunctionTransformer", calculate_sklearn_function_transformer_output_shapes
)
